# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from unittest import mock

from neutron.tests.unit.plugins.ml2 import test_plugin
from neutron_lib import context
from oslo_config import cfg
from oslo_log import log as logging
import webob.exc

from gbpservice.neutron.db.grouppolicy import group_policy_mapping_db as gpmdb
from gbpservice.neutron.extensions import group_policy as gpolicy
from gbpservice.neutron.services.grouppolicy import config
from gbpservice.neutron.services.grouppolicy.drivers import dummy_driver
from gbpservice.neutron.services.grouppolicy import plugin as gplugin
from gbpservice.neutron.tests.unit.db.grouppolicy import (
    test_group_policy_db as tgpdb)
from gbpservice.neutron.tests.unit.db.grouppolicy import (
    test_group_policy_mapping_db as tgpmdb)


LOG = logging.getLogger(__name__)
cfg.CONF.import_opt('policy_drivers',
                    'gbpservice.neutron.services.grouppolicy.config',
                    group='group_policy')
GP_PLUGIN_KLASS = (
    "gbpservice.neutron.services.grouppolicy.plugin.GroupPolicyPlugin"
)
SERVICECHAIN_SPECS = 'servicechain/servicechain_specs'
SERVICECHAIN_NODES = 'servicechain/servicechain_nodes'


class FakeDriver(object):

    def _fill_order(self, context):
        context.call_order.append(self)

    def __getattr__(self, item):
        return self._fill_order


NEW_STATUS = 'new_status'
NEW_STATUS_DETAILS = 'new_status_details'


def get_status_for_test(self, context):
    resource_name = [item for item in list(context.__dict__.keys())
     if item.startswith('_original')][0][len('_original'):]
    getattr(context, resource_name)['status'] = NEW_STATUS
    getattr(context, resource_name)['status_details'] = NEW_STATUS_DETAILS


class GroupPolicyPluginTestBase(tgpmdb.GroupPolicyMappingDbTestCase):

    def setUp(self, core_plugin=None, l3_plugin=None, gp_plugin=None,
              ml2_options=None, sc_plugin=None, qos_plugin=None,
              trunk_plugin=None):
        if not gp_plugin:
            gp_plugin = GP_PLUGIN_KLASS
        ml2_opts = ml2_options or {'mechanism_drivers': ['openvswitch'],
                                   'extension_drivers': ['port_security']}
        for opt, val in list(ml2_opts.items()):
            cfg.CONF.set_override(opt, val, 'ml2')
        core_plugin = core_plugin or test_plugin.PLUGIN_NAME
        super(GroupPolicyPluginTestBase, self).setUp(core_plugin=core_plugin,
                                                     l3_plugin=l3_plugin,
                                                     gp_plugin=gp_plugin,
                                                     sc_plugin=sc_plugin,
                                                     qos_plugin=qos_plugin,
                                                     trunk_plugin=trunk_plugin)

    def _create_l2_policy_on_shared(self, **kwargs):
        l3p = self.create_l3_policy(shared=True)['l3_policy']
        return self.create_l2_policy(l3_policy_id=l3p['id'],
                                     **kwargs)['l2_policy']

    def _create_ptg_on_shared(self, **kwargs):
        l2p = self._create_l2_policy_on_shared(shared=True)
        return self.create_policy_target_group(l2_policy_id=l2p['id'],
                                               **kwargs)

    def _create_rule_on_shared(self, **kwargs):
        pa = self.create_policy_action(action_type='allow',
                                       shared=True)['policy_action']
        cl_attr = {'protocol': 'tcp', 'port_range': 80}
        pc = self.create_policy_classifier(direction='in',
                                           shared=True,
                                           **cl_attr)['policy_classifier']
        return self.create_policy_rule(
            policy_classifier_id=pc['id'],
            policy_actions=[pa['id']], **kwargs)['policy_rule']

    def _create_policy_rule_set_on_shared(self, **kwargs):
        pr = self._create_rule_on_shared(shared=True)
        return self.create_policy_rule_set(policy_rules=[pr['id']],
                                           **kwargs)['policy_rule_set']

    def _create_external_policy_on_shared(self, **kwargs):
        es = self.create_external_segment(shared=True)
        return self.create_external_policy(
            external_segments=[es['external_segment']['id']],
            **kwargs)['external_policy']

    def _create_nat_pool_on_shared(self, **kwargs):
        es = self.create_external_segment(shared=True)
        return self.create_nat_pool(
            external_segment_id=es['external_segment']['id'],
            **kwargs)['nat_pool']

    def _create_servicechain_spec(self, node_types=None, shared=False):
        node_types = node_types or []
        if not node_types:
            node_types = ['LOADBALANCERV2']
        node_ids = []
        for node_type in node_types:
            node_ids.append(self._create_servicechain_node(node_type,
                                                           shared=shared))
        data = {'servicechain_spec': {'tenant_id': self._tenant_id if not
                                      shared else 'another-tenant',
                                      'nodes': node_ids,
                                      'shared': shared}}
        scs_req = self.new_create_request(
            SERVICECHAIN_SPECS, data, self.fmt)
        spec = self.deserialize(
            self.fmt, scs_req.get_response(self.ext_api))
        scs_id = spec['servicechain_spec']['id']
        return scs_id

    def _create_servicechain_node(self, node_type="LOADBALANCERV2",
                                  shared=False):
        config = "{}"
        data = {'servicechain_node': {'service_type': node_type,
                                      'tenant_id': self._tenant_id if not
                                      shared else 'another-tenant',
                                      'config': config,
                                      'shared': shared}}
        scn_req = self.new_create_request(SERVICECHAIN_NODES, data, self.fmt)
        node = self.deserialize(self.fmt, scn_req.get_response(self.ext_api))
        scn_id = node['servicechain_node']['id']
        return scn_id

    def _get_object(self, type, id, api, expected_res_status=None):
        req = self.new_show_request(type, id, self.fmt)
        res = req.get_response(api)

        if expected_res_status:
            self.assertEqual(res.status_int, expected_res_status)
        elif res.status_int >= webob.exc.HTTPClientError.code:
            raise webob.exc.HTTPClientError(code=res.status_int)

        return self.deserialize(self.fmt, res)


class GroupPolicyPluginTestCase(GroupPolicyPluginTestBase):

    def tearDown(self):
        policy_drivers = cfg.CONF.group_policy.policy_drivers
        LOG.warning("PDs used in this test: %s",
                    policy_drivers)
        # Always reset configuration to dummy driver. Any
        # test which requires to configure a different
        # policy driver would have done so in it's setup
        # (and should have ideally reset it too).
        config.cfg.CONF.set_override('policy_drivers',
                                     ['dummy'],
                                     group='group_policy')
        super(GroupPolicyPluginTestCase, self).tearDown()


class TestGroupPolicyPluginDrivers(GroupPolicyPluginTestBase):

    def test_reverse_on_delete(self):
        manager = self.plugin.policy_driver_manager
        ctx = context.get_admin_context()
        drivers = manager.ordered_policy_drivers
        first, second = mock.Mock(), mock.Mock()
        first.obj, second.obj = FakeDriver(), FakeDriver()
        try:
            manager.ordered_policy_drivers = [first, second]
            manager.reverse_ordered_policy_drivers = [second, first]
            ordered_obj = [first.obj, second.obj]
            ctx.call_order = []
            manager._call_on_drivers('nodelete', ctx)
            self.assertEqual(ordered_obj, ctx.call_order)
            ctx.call_order = []
            manager._call_on_drivers('delete', ctx)
            self.assertEqual(ordered_obj[::-1], ctx.call_order)
        finally:
            manager.ordered_policy_drivers = drivers


class TestL3Policy(GroupPolicyPluginTestCase):

    def _get_es_dict(self, es, addr=None):
        return {es['external_segment']['id']: addr or []}

    def test_shared_l3_policy_create(self):
        # Verify default is False
        l3p = self.create_l3_policy()
        self.assertFalse(l3p['l3_policy']['shared'])
        # Verify shared True created without errors
        l3p = self.create_l3_policy(shared=True)
        self.assertTrue(l3p['l3_policy']['shared'])

    def test_shared_l3p_create_with_es(self):
        def combination(l3p, es):
            return {'l3p': l3p, 'es': es}
        allowed = [combination(False, False), combination(True, True),
                   combination(False, True)]
        for shared in allowed:
            es = self.create_external_segment(
                cidr='172.0.0.0/8', shared=shared['es'])
            es_dict = self._get_es_dict(es, ['172.0.0.2', '172.0.0.3'])
            l3p = self.create_l3_policy(
                external_segments=es_dict, shared=shared['l3p'],
                expected_res_status=201)['l3_policy']
            # Verify create successful
            self.assertEqual(es_dict, l3p['external_segments'])

    def test_shared_l3p_create_with_es_negative(self):
        # Not allowed: Unshared ES with shared L3P
        es = self.create_external_segment(cidr='172.0.0.0/8')
        es_dict = self._get_es_dict(es, ['172.0.0.2', '172.0.0.3'])
        res = self.create_l3_policy(external_segments=es_dict,
                                    shared=True,
                                    expected_res_status=400)
        self.assertEqual('SharedResourceReferenceError',
                         res['NeutronError']['type'])

    def test_shared_l3_policy_update(self):
        l3p = self.create_l3_policy()['l3_policy']
        # Accept share if nothing referenced
        self.update_l3_policy(l3p['id'], expected_res_status=200, shared=True)

        # Verify unshare when referenced by private L2P
        self.create_l2_policy(l3_policy_id=l3p['id'])
        self.update_l3_policy(l3p['id'], expected_res_status=200, shared=False)
        es = self.create_external_segment(cidr='172.0.0.0/8')
        es_dict = self._get_es_dict(es, ['172.0.0.2', '172.0.0.3'])
        # Set ES
        l3p = self.update_l3_policy(l3p['id'], expected_res_status=200,
                                    external_segments=es_dict)['l3_policy']
        self.assertEqual(es_dict, l3p['external_segments'])

        # Share ES
        self.update_external_segment(
            es['external_segment']['id'], expected_res_status=200, shared=True)

        # Verify sharing/unsharing successful
        for shared in [True, False]:
            self.update_l3_policy(l3p['id'],
                                  expected_res_status=200, shared=shared)

        # Remove ES
        l3p = self.update_l3_policy(l3p['id'], expected_res_status=200,
                                    external_segments={})['l3_policy']
        self.assertEqual({}, l3p['external_segments'])
        # Verify ES update with sharing successful
        l3p = self.update_l3_policy(
            l3p['id'], expected_res_status=200, external_segments=es_dict,
            shared=True)['l3_policy']
        # Verify ES correctly set
        self.assertEqual(es_dict, l3p['external_segments'])

    def test_shared_l3_policy_update_negative(self):
        l3p = self.create_l3_policy(shared=True)['l3_policy']
        self.create_l2_policy(l3_policy_id=l3p['id'], shared=True)
        # Unshare not possible when reference by shared resource
        self.update_l3_policy(l3p['id'], expected_res_status=400, shared=False)

        l3p = self.create_l3_policy(shared=True)['l3_policy']
        self.create_l2_policy(l3_policy_id=l3p['id'], shared=False,
                              tenant_id='other')
        # Unshare not possible when referenced by other tenant's
        # private resource
        self.update_l3_policy(l3p['id'], expected_res_status=400, shared=False)

        es = self.create_external_segment(cidr='172.0.0.0/8')
        es_dict = self._get_es_dict(es, ['172.0.0.2', '172.0.0.3'])
        res = self.update_l3_policy(
            l3p['id'], expected_res_status=400,
            external_segments=es_dict, shared=True)
        self.assertEqual('SharedResourceReferenceError',
                         res['NeutronError']['type'])

    def test_create_with_es_negative(self):
        attrs = {'external_routes': [{'destination': '10.160.0.0/16',
                                     'nexthop': '172.1.1.1'}],
                 'cidr': '172.1.1.0/24'}
        es = self.create_external_segment(**attrs)['external_segment']
        # Overlapping pool
        attrs = {'ip_pool': '172.1.1.0/20',
                 'external_segments': {es['id']: ['172.1.1.2']}}
        res = self.create_l3_policy(expected_res_status=400, **attrs)
        self.assertEqual('ExternalSegmentSubnetOverlapsWithL3PIpPool',
                         res['NeutronError']['type'])
        # Overlapping route
        attrs['ip_pool'] = '10.160.1.0/24'
        res = self.create_l3_policy(expected_res_status=400, **attrs)
        self.assertEqual('ExternalRouteOverlapsWithL3PIpPool',
                         res['NeutronError']['type'])
        # Allocated address not in pool
        attrs = {'ip_pool': '192.168.0.0/24',
                 'external_segments': {es['id']: ['172.1.2.2']}}
        res = self.create_l3_policy(expected_res_status=400, **attrs)
        self.assertEqual('InvalidL3PExternalIPAddress',
                         res['NeutronError']['type'])

    def test_update_with_es_negative(self):
        attrs = {'external_routes': [{'destination': '10.160.0.0/16',
                                     'nexthop': '172.1.1.1'}],
                 'cidr': '172.1.1.0/24'}
        es = self.create_external_segment(**attrs)['external_segment']

        # Overlapping pool
        l3p = self.create_l3_policy(ip_pool='172.1.1.0/20')['l3_policy']
        attrs = {'external_segments': {es['id']: ['172.1.1.2']}}
        res = self.update_l3_policy(
            l3p['id'], expected_res_status=400, **attrs)
        self.assertEqual('ExternalSegmentSubnetOverlapsWithL3PIpPool',
                         res['NeutronError']['type'])

        # Overlapping route
        l3p = self.create_l3_policy(ip_pool='10.160.1.0/24')['l3_policy']
        res = self.update_l3_policy(
            l3p['id'], expected_res_status=400, **attrs)
        self.assertEqual('ExternalRouteOverlapsWithL3PIpPool',
                         res['NeutronError']['type'])

        # Allocated address not in pool
        l3p = self.create_l3_policy(ip_pool='192.168.0.0/24')['l3_policy']
        attrs = {'external_segments': {es['id']: ['172.1.2.2']}}
        res = self.update_l3_policy(
            l3p['id'], expected_res_status=400, **attrs)
        self.assertEqual('InvalidL3PExternalIPAddress',
                         res['NeutronError']['type'])

    def test_multiple_es_negative(self):
        attrs = {'external_routes': [{'destination': '0.0.0.0/0',
                                      'nexthop': '172.1.1.1'}],
                 'cidr': '172.1.1.0/24'}
        es = self.create_external_segment(**attrs)['external_segment']
        another_es = self.create_external_segment(**attrs)['external_segment']

        res = self.create_l3_policy(
            external_segments={es['id']: [], another_es['id']: []},
            expected_res_status=400)
        self.assertEqual('IdenticalExternalRoute',
                         res['NeutronError']['type'])

        l3p = self.create_l3_policy(
            external_segments={es['id']: []},
            expected_res_status=201)['l3_policy']
        res = self.update_l3_policy(l3p['id'],
            external_segments={es['id']: [], another_es['id']: []},
            expected_res_status=400)
        self.assertEqual('IdenticalExternalRoute',
                         res['NeutronError']['type'])


class TestL2Policy(GroupPolicyPluginTestCase):

    def test_shared_l2_policy_create(self):
        l3p = self.create_l3_policy(shared=True)['l3_policy']
        # Verify Default False
        l2p = self.create_l2_policy(l3_policy_id=l3p['id'])
        self.assertFalse(l2p['l2_policy']['shared'])
        # Verify shared True created without errors
        l2p = self.create_l2_policy(l3_policy_id=l3p['id'], shared=True)
        self.assertTrue(l2p['l2_policy']['shared'])

    def test_shared_l2_policy_update(self):
        l2p = self._create_l2_policy_on_shared()
        self.update_l2_policy(l2p['id'], expected_res_status=200, shared=True)

        self.create_policy_target_group(l2_policy_id=l2p['id'])
        self.update_l2_policy(l2p['id'], expected_res_status=200, shared=False)

        # Verify l2p can be moved across shared l3p
        l2p = self._create_l2_policy_on_shared(
            shared=True)
        l3p = self.create_l3_policy(
            shared=True)['l3_policy']
        self.update_l2_policy(l2p['id'], expected_res_status=200,
                              l3_policy_id=l3p['id'])

    def test_shared_l2_policy_create_negative(self):
        l3p = self.create_l3_policy()['l3_policy']
        self.create_l2_policy(l3_policy_id=l3p['id'], shared=True,
                              expected_res_status=400)

        # Verify shared L2p can't be moved to a non shared L3p
        l2p = self._create_l2_policy_on_shared(shared=True)
        l3p = self.create_l3_policy()['l3_policy']
        self.update_l2_policy(l2p['id'], expected_res_status=400,
                              l3_policy_id=l3p['id'])

    def test_shared_l2_policy_update_negative(self):
        l2p = self._create_l2_policy_on_shared(shared=True)
        self.create_policy_target_group(l2_policy_id=l2p['id'], shared=True)
        self.update_l2_policy(l2p['id'], expected_res_status=400, shared=False)

        l2p = self._create_l2_policy_on_shared(shared=True)
        self.create_policy_target_group(
            l2_policy_id=l2p['id'], shared=False, tenant_id='other')
        self.update_l2_policy(l2p['id'], expected_res_status=400, shared=False)

    def test_l2p_create_among_tenants(self):
        # L2P on shared L3P:
        self._create_l2_policy_on_shared(
            tenant_id='other', expected_res_status=201)


class TestPolicyRuleSet(GroupPolicyPluginTestCase):

    def test_shared_policy_rule_set_create(self):
        # Verify shared policy_rule_set created with shared rules
        prs = self._create_policy_rule_set_on_shared(
            shared=True, expected_res_status=201)
        self.assertTrue(prs['shared'])

        # Verify non shared policy_rule_set created with shared rules
        prs = self._create_policy_rule_set_on_shared(expected_res_status=201)
        self.assertFalse(prs['shared'])

    def test_shared_policy_rule_set_update(self):
        prs = self._create_policy_rule_set_on_shared()
        self.update_policy_rule_set(prs['id'],
                                    expected_res_status=200, shared=True)

        self.create_policy_target_group(
            provided_policy_rule_sets={prs['id']: None})
        self.update_policy_rule_set(
            prs['id'], expected_res_status=200, shared=False)

    def test_shared_policy_rule_set_create_negative(self):
        # Verify shared policy_rule_set fails with non shared rules
        prns = self._create_rule_on_shared()
        pr = self._create_rule_on_shared(shared=True)
        self.create_policy_rule_set(expected_res_status=400,
                                    shared=True,
                                    policy_rules=[pr['id'], prns['id']])

    def test_shared_policy_rule_set_update_negative(self):
        prs = self._create_policy_rule_set_on_shared(shared=True)
        self.create_policy_target_group(
            provided_policy_rule_sets={prs['id']: None}, shared=True)
        self.update_policy_rule_set(
            prs['id'], expected_res_status=400, shared=False)

        prs = self._create_policy_rule_set_on_shared(shared=True)
        self.create_policy_target_group(
            provided_policy_rule_sets={prs['id']: None},
            shared=False, tenant_id='other')
        self.update_policy_rule_set(
            prs['id'], expected_res_status=400, shared=False)
        # Verify non shared rules can't be set on non shared prs
        nsr = self._create_rule_on_shared()
        self.update_policy_rule_set(prs['id'], expected_res_status=400,
                                    policy_rules=[nsr['id']])

    def test_policy_rule_set_create_among_tenants(self):
        self._create_policy_rule_set_on_shared(tenant_id='other',
                                               expected_res_status=201)


class TestPolicyRule(GroupPolicyPluginTestCase):

    def test_shared_rule_create(self):
        # Verify shared rule created with shared actions and classifier
        pr = self._create_rule_on_shared(shared=True,
                                         expected_res_status=201)
        self.assertTrue(pr['shared'])

        # Verify non shared rule create with shared actions and classifier
        pr = self._create_rule_on_shared(expected_res_status=201)
        self.assertFalse(pr['shared'])

    def test_shared_rule_update(self):
        pr = self._create_rule_on_shared()
        self.update_policy_rule(pr['id'], expected_res_status=200, shared=True)

        self.create_policy_rule_set(policy_rules=[pr['id']])
        self.update_policy_rule(
            pr['id'], expected_res_status=200, shared=False)

    def test_shared_rule_create_negative(self):
        # Verify shared rule fails with non shared classifier
        pans = self.create_policy_action(action_type='allow')['policy_action']
        cl_attr = {'protocol': 'tcp', 'port_range': 80}
        pcns = self.create_policy_classifier(
            direction='in', **cl_attr)['policy_classifier']
        pc = self.create_policy_classifier(
            direction='in', shared=True,
            **cl_attr)['policy_classifier']
        self.create_policy_rule(
            policy_classifier_id=pcns['id'], expected_res_status=400,
            shared=True)

        # Verify shared rule fails with non shared action
        self.create_policy_rule(
            policy_classifier_id=pc['id'], policy_actions=[pans['id']],
            expected_res_status=400, shared=True)

    def test_shared_rule_update_negative(self):
        pr = self._create_rule_on_shared(shared=True)
        self.create_policy_rule_set(policy_rules=[pr['id']], shared=True,
                                    tenant_id='another')
        self.update_policy_rule(pr['id'],
                                expected_res_status=400, shared=False)

        pr = self._create_rule_on_shared(shared=True)
        self.create_policy_rule_set(policy_rules=[pr['id']], shared=False,
                                    tenant_id='other')
        self.update_policy_rule(pr['id'],
                                expected_res_status=400, shared=False)

    def test_rule_create_among_tenants(self):
        self._create_rule_on_shared(tenant_id='other',
                                    expected_res_status=201)


class TestPolicyClassifier(GroupPolicyPluginTestCase):
    def test_shared_policy_classifier_update(self):
        cl_attr = {'protocol': 'tcp', 'port_range': 80}
        pc = self.create_policy_classifier(**cl_attr)['policy_classifier']
        pa = self.create_policy_action(action_type='allow')['policy_action']
        self.update_policy_classifier(
            pc['id'], expected_res_status=200, shared=True)

        self.create_policy_rule(policy_classifier_id=pc['id'],
                                policy_actions=[pa['id']])
        self.update_policy_classifier(pc['id'],
                                      expected_res_status=200, shared=False)

    def test_shared_policy_classifier_update_negative(self):
        cl_attr = {'protocol': 'tcp', 'port_range': 80}
        pc = self.create_policy_classifier(shared=True,
                                           **cl_attr)['policy_classifier']
        pa = self.create_policy_action(
            action_type='allow', shared=True)['policy_action']
        self.create_policy_rule(
            policy_classifier_id=pc['id'], policy_actions=[pa['id']],
            shared=True)
        self.update_policy_classifier(pc['id'],
                                      expected_res_status=400, shared=False)

        self.create_policy_rule(policy_classifier_id=pc['id'],
                                policy_actions=[pa['id']], shared=False,
                                tenant_id='other')
        self.update_policy_classifier(pc['id'],
                                      expected_res_status=400, shared=False)


class TestPolicyTargetGroup(GroupPolicyPluginTestCase):

    def test_delete_ptg_with_unused_pt(self):
        ctx = context.get_admin_context()
        ptg = self.create_policy_target_group()['policy_target_group']
        self.create_policy_target(policy_target_group_id=ptg['id'])
        req = self.new_delete_request('policy_target_groups', ptg['id'],
                                      self.fmt)
        res = req.get_response(self.ext_api)
        self.assertEqual(webob.exc.HTTPNoContent.code, res.status_int)
        self.assertRaises(gpolicy.PolicyTargetGroupNotFound,
                          self.plugin.get_policy_target_group, ctx, ptg['id'])

    def test_delete_fails_on_used_ptg(self):
        with self.port() as port:
            port_id = port['port']['id']
            ptg = self.create_policy_target_group()['policy_target_group']
            pt = self.create_policy_target(policy_target_group_id=ptg['id'],
                                      port_id=port_id)
            port = self._get_object('ports', pt['policy_target']['port_id'],
                                    self.api)
            self._bind_port_to_host(port['port']['id'], 'h1')
            self.delete_policy_target_group(ptg['id'],
                                            expected_res_status=400)
            self._unbind_port(port['port']['id'])
            self.delete_policy_target_group(ptg['id'], expected_res_status=204)

    def test_shared_ptg_create(self):
        l2p = self._create_l2_policy_on_shared(
            shared=True)
        l2pns = self._create_l2_policy_on_shared()

        prs = self._create_policy_rule_set_on_shared(
            shared=True)
        ctns = self._create_policy_rule_set_on_shared()

        nsp = self.create_network_service_policy(
            shared=True)['network_service_policy']
        nspns = self.create_network_service_policy()['network_service_policy']
        # Verify non-shared ptg providing and consuming shared and non shared
        # policy_rule_sets
        ptg = self.create_policy_target_group(
            l2_policy_id=l2p['id'], expected_res_status=201)
        self.assertFalse(ptg['policy_target_group']['shared'])
        ptg = self.create_policy_target_group(
            l2_policy_id=l2p['id'],
            provided_policy_rule_sets={prs['id']: '', ctns['id']: ''},
            consumed_policy_rule_sets={prs['id']: '', ctns['id']: ''},
            expected_res_status=201)
        self.assertFalse(ptg['policy_target_group']['shared'])
        ptg = self.create_policy_target_group(
            l2_policy_id=l2p['id'], network_service_policy_id=nsp['id'],
            expected_res_status=201)
        self.assertFalse(ptg['policy_target_group']['shared'])
        ptg = self.create_policy_target_group(
            l2_policy_id=l2p['id'], network_service_policy_id=nspns['id'],
            expected_res_status=201)
        self.assertFalse(ptg['policy_target_group']['shared'])

        # Verify shared True created without errors by providing/consuming
        # shared policy_rule_sets
        ptg = self.create_policy_target_group(
            l2_policy_id=l2p['id'], shared=True,
            expected_res_status=201)
        self.assertTrue(ptg['policy_target_group']['shared'])
        ptg = self.create_policy_target_group(
            l2_policy_id=l2p['id'], provided_policy_rule_sets={prs['id']: ''},
            consumed_policy_rule_sets={prs['id']: ''}, shared=True,
            expected_res_status=201)
        self.assertTrue(ptg['policy_target_group']['shared'])

        # Verify not shared created without error on not shared l2p
        self.create_policy_target_group(l2_policy_id=l2pns['id'],
                                   expected_res_status=201)

    def test_shared_ptg_update(self):
        ptg = self._create_ptg_on_shared()['policy_target_group']
        self.update_policy_target_group(
            ptg['id'], expected_res_status=200, shared=True)

        self.create_policy_target(policy_target_group_id=ptg['id'])
        self.update_policy_target_group(
            ptg['id'], expected_res_status=200, shared=False)

    def test_shared_ptg_create_negative(self):
        l2pns = self._create_l2_policy_on_shared()
        ctns = self._create_policy_rule_set_on_shared()
        # Verify shared PTG fails on non-shared l2p
        self.create_policy_target_group(
            l2_policy_id=l2pns['id'], shared=True,
            expected_res_status=400)
        # Verify shared PTG fails to provide/consume non shared
        # policy_rule_sets
        self._create_ptg_on_shared(
            shared=True,
            provided_policy_rule_sets={ctns['id']: ''},
            consumed_policy_rule_sets={ctns['id']: ''},
            expected_res_status=400)

    def test_shared_ptg_update_negative(self):
        ptg = self._create_ptg_on_shared(
            shared=True)['policy_target_group']
        self.create_policy_target(policy_target_group_id=ptg['id'],
                                  tenant_id='other')
        self.update_policy_target_group(
            ptg['id'], expected_res_status=400, shared=False)

        # Verify update to non shared L2p fails
        l2p = self.create_l2_policy()['l2_policy']
        self.update_policy_target_group(
            ptg['id'], expected_res_status=400,
            l2_policy_id=l2p['id'])

        # Verify update to non shared NSP fails
        nsp = self.create_network_service_policy()['network_service_policy']
        self.update_policy_target_group(
            ptg['id'], expected_res_status=400,
            network_service_policy_id=nsp['id'])

        # Verify update to non shared provided PRS fails
        pts = self._create_policy_rule_set_on_shared()
        self.update_policy_target_group(
            ptg['id'], expected_res_status=400,
            provided_policy_rule_sets={pts['id']: ''})
        # Verify update to non shared consumed PRS fails
        self.update_policy_target_group(
            ptg['id'], expected_res_status=400,
            consumed_policy_rule_sets={pts['id']: ''})

    def test_complex_ptg_create_among_tenant(self):
        ctp = self._create_policy_rule_set_on_shared(
            shared=True)
        ctc = self._create_policy_rule_set_on_shared(
            shared=True)
        nsp = self.create_network_service_policy(
            shared=True)['network_service_policy']
        self._create_ptg_on_shared(
            tenant_id='other', provided_policy_rule_sets={ctp['id']: ''},
            consumed_policy_rule_sets={ctc['id']: ''},
            network_service_policy_id=nsp['id'], expected_res_status=201)

    def test_ptg_create_among_tenants(self):
        self._create_ptg_on_shared(tenant_id='other',
                                   expected_res_status=201)

    def test_multiple_service_ptg_fails(self):
        self.create_policy_target_group(
            service_management=True, is_admin_context=True,
            expected_res_status=201)
        res = self.create_policy_target_group(
            service_management=True, is_admin_context=True,
            expected_res_status=400)
        self.assertEqual('ManagementPolicyTargetGroupExists',
                         res['NeutronError']['type'])

    def test_update_l2p_rejectet(self):
        l2p_1 = self.create_l2_policy()['l2_policy']
        l2p_2 = self.create_l2_policy()['l2_policy']
        ptg = self.create_policy_target_group(
            l2_policy_id=l2p_1['id'],
            expected_res_status=201)['policy_target_group']
        res = self.update_policy_target_group(
            ptg['id'], l2_policy_id=l2p_2['id'], expected_res_status=400)
        self.assertEqual('L2PolicyUpdateOfPolicyTargetGroupNotSupported',
                         res['NeutronError']['type'])


class TestExternalSegment(GroupPolicyPluginTestCase):

    def test_shared_es_create(self):
        # Verify default is False
        es = self.create_external_segment()
        self.assertFalse(es['external_segment']['shared'])
        # Verify shared True created without errors
        es = self.create_external_segment(shared=True)
        self.assertTrue(es['external_segment']['shared'])

    def test_shared_es_update(self):
        es = self.create_external_segment()['external_segment']
        for shared in [True, False]:
            self.update_external_segment(
                es['id'], expected_res_status=200,
                shared=shared)

    def test_create_routes(self):
        attrs = {'external_routes': [{'destination': '0.0.0.0/0',
                                     'nexthop': '172.1.0.1'}],
                 'cidr': '172.1.0.0/24'}
        self.create_external_segment(expected_res_status=201, **attrs)

    def test_routes_negative(self):
        # Verify wrong NH
        attrs = {'external_routes': [{'destination': '0.0.0.0/0',
                                     'nexthop': '172.1.1.1'}],
                 'cidr': '172.1.0.0/24'}
        res = self.create_external_segment(expected_res_status=400, **attrs)
        self.assertEqual('ExternalRouteNextHopNotInExternalSegment',
                         res['NeutronError']['type'])
        attrs['cidr'] = '172.1.1.0/24'
        es = self.create_external_segment(**attrs)['external_segment']
        self.create_l3_policy(
            ip_pool='192.160.0.0/16',
            external_segments={es['id']: ['172.1.1.2']})['l3_policy']

        # Verify refused because overlapping with L3P
        attrs = {'external_routes': [{'destination': '192.168.2.0/0',
                                     'nexthop': '172.1.1.1'}]}
        res = self.update_external_segment(
            es['id'], expected_res_status=400, **attrs)
        self.assertEqual('ExternalRouteOverlapsWithL3PIpPool',
                         res['NeutronError']['type'])

    def test_identical_routes_negative(self):
        attrs = {'external_routes': [{'destination': '0.0.0.0/0',
                                      'nexthop': '172.1.0.1'}],
                 'cidr': '172.1.0.0/24'}
        es1 = self.create_external_segment(**attrs)['external_segment']
        es2 = self.create_external_segment()['external_segment']

        self.create_l3_policy(
            external_segments={es1['id']: [], es2['id']: []},
            expected_res_status=201)

        del attrs['cidr']
        res = self.update_external_segment(
            es2['id'], expected_res_status=400, **attrs)
        self.assertEqual('IdenticalExternalRoute',
                         res['NeutronError']['type'])


class TestExternalPolicy(GroupPolicyPluginTestCase):

    def test_shared_ep_create(self):
        es = self.create_external_segment(
            shared=True)['external_segment']
        esns = self.create_external_segment(
        )['external_segment']

        prs = self._create_policy_rule_set_on_shared(shared=True)
        prsns = self._create_policy_rule_set_on_shared()

        # Verify non-shared ep providing and consuming shared and non shared
        # policy_rule_sets
        ep = self.create_external_policy(
            external_segments=[es['id']], expected_res_status=201)
        self.assertFalse(ep['external_policy']['shared'])
        ep = self.create_external_policy(
            external_segments=[es['id']],
            provided_policy_rule_sets={prs['id']: '', prsns['id']: ''},
            consumed_policy_rule_sets={prs['id']: '', prsns['id']: ''},
            expected_res_status=201)
        self.assertFalse(ep['external_policy']['shared'])

        # Verify shared True created without errors by providing/consuming
        # shared policy_rule_sets
        ep = self.create_external_policy(
            external_segments=[es['id']], shared=True,
            expected_res_status=201)
        self.assertTrue(ep['external_policy']['shared'])
        ep = self.create_external_policy(
            external_segments=[es['id']],
            provided_policy_rule_sets={prs['id']: ''},
            consumed_policy_rule_sets={prs['id']: ''}, shared=True,
            expected_res_status=201)
        self.assertTrue(ep['external_policy']['shared'])

        # Verify not shared created without error on not shared es
        self.create_external_policy(
            external_segments=[esns['id']], expected_res_status=201)

    def test_shared_ep_update(self):
        ep = self._create_external_policy_on_shared()
        self.update_external_policy(
            ep['id'], expected_res_status=200, shared=True)
        self.update_external_policy(
            ep['id'], expected_res_status=200, shared=False)

    def test_shared_ep_create_negative(self):
        es = self.create_external_segment()['external_segment']
        prs = self._create_policy_rule_set_on_shared()
        # Verify shared EP fails on non-shared es
        res = self.create_external_policy(
            external_segments=[es['id']], shared=True,
            expected_res_status=400)
        self.assertEqual('SharedResourceReferenceError',
                         res['NeutronError']['type'])
        # Verify shared EP fails to provide/consume non shared
        # policy_rule_sets
        res = self.create_external_policy(
            shared=True,
            provided_policy_rule_sets={prs['id']: ''},
            consumed_policy_rule_sets={prs['id']: ''},
            expected_res_status=400)
        self.assertEqual('SharedResourceReferenceError',
                         res['NeutronError']['type'])

    def test_shared_ep_update_negative(self):
        ep = self._create_external_policy_on_shared(shared=True)
        # Verify update to non shared ES fails
        es = self.create_external_segment()['external_segment']
        self.update_external_policy(
            ep['id'], expected_res_status=400, external_segments=[es['id']])

        # Verify update to non shared provided PRS fails
        prs = self._create_policy_rule_set_on_shared()
        self.update_external_policy(
            ep['id'], expected_res_status=400,
            provided_policy_rule_sets={prs['id']: ''})
        # Verify update to non shared consumed PRS fails
        self.update_external_policy(
            ep['id'], expected_res_status=400,
            consumed_policy_rule_sets={prs['id']: ''})


class TestNatPool(GroupPolicyPluginTestCase):

    def test_nat_pool_shared_create(self):
        def combination(np, es):
            return {'np': np, 'es': es}
        allowed = [combination(False, False), combination(True, True),
                   combination(False, True)]
        for shared in allowed:
            es = self.create_external_segment(
                shared=shared['es'])['external_segment']
            self.create_nat_pool(
                external_segment_id=es['id'], shared=shared['np'],
                expected_res_status=201)

    def test_nat_pool_shared_create_negative(self):
        es = self.create_external_segment(
            shared=False)['external_segment']
        res = self.create_nat_pool(
            external_segment_id=es['id'], shared=True,
            expected_res_status=400)
        self.assertEqual('SharedResourceReferenceError',
                         res['NeutronError']['type'])

    def test_nat_pool_shared_update(self):
        np = self.create_nat_pool(shared=False)['nat_pool']
        for shared in [False, True]:
            es = self.create_external_segment(
                shared=shared)['external_segment']
            self.update_nat_pool(
                np['id'], expected_res_status=200,
                external_segment_id=es['id'])
        np = self.create_nat_pool(shared=True)['nat_pool']
        es = self.create_external_segment(
            shared=True)['external_segment']
        # Verify shared NP on shared ES
        self.update_nat_pool(
            np['id'], expected_res_status=200,
            external_segment_id=es['id'])
        # Verify unshare NP
        self.update_nat_pool(
            np['id'], expected_res_status=200, shared=False)


class TestPolicyTarget(GroupPolicyPluginTestCase):

    def _test_cross_tenant(self, is_admin=False):
        status = {False: 404, True: 201}
        ptg = self.create_policy_target_group(
            expected_res_status=201, tenant_id='tenant',
            is_admin_context=is_admin)['policy_target_group']
        # Create EP on a different tenant
        res = self.create_policy_target(
            expected_res_status=status[is_admin], tenant_id='another',
            is_admin_context=is_admin, policy_target_group_id=ptg['id'])
        if not is_admin:
            self.assertEqual(
                'GbpResourceNotFound', res['NeutronError']['type'])

        # Create EP without PTG
        pt = self.create_policy_target(
            expected_res_status=201, tenant_id='another',
            is_admin_context=is_admin)['policy_target']

        # Update PT fails
        res = self.update_policy_target(
            pt['id'], tenant_id='another', policy_target_group_id=ptg['id'],
            expected_res_status=status[is_admin] if not is_admin else 200,
            is_admin_context=is_admin)

        if not is_admin:
            self.assertEqual(
                'GbpResourceNotFound', res['NeutronError']['type'])

    def test_cross_tenant_fails(self):
        self._test_cross_tenant()

    def test_cross_tenant_admin(self):
        self._test_cross_tenant(True)


class TestResourceStatusChange(GroupPolicyPluginTestCase):

    def setUp(self, core_plugin=None, gp_plugin=None, ml2_options=None,
              sc_plugin=None):
        for resource_name in gpolicy.RESOURCE_ATTRIBUTE_MAP:
            method_name = "get_" + self._get_resource_singular(
                resource_name) + "_status"
            setattr(dummy_driver.NoopDriver, method_name, get_status_for_test)
        super(TestResourceStatusChange, self).setUp(
            core_plugin=core_plugin, gp_plugin=gp_plugin, sc_plugin=sc_plugin)

    def _test_status_change_on_get(self, resource_name, fields=None):
        resource_singular = self._get_resource_singular(resource_name)
        if resource_name == 'policy_rules':
            pc_id = self.create_policy_classifier()['policy_classifier']['id']
            resource = self.create_policy_rule(policy_classifier_id=pc_id)
        else:
            resource = getattr(self, "create_" + resource_singular)()
        self.assertEqual(NEW_STATUS, resource[resource_singular]['status'])
        self.assertEqual(NEW_STATUS_DETAILS,
                         resource[resource_singular]['status_details'])

        # Reset status directly in the DB to test that GET works
        reset_status = {resource_singular: {'status': None,
                                            'status_details': None}}
        neutron_context = context.Context('', self._tenant_id)
        getattr(gpmdb.GroupPolicyMappingDbPlugin,
                "update_" + resource_singular)(
                    self._gbp_plugin, neutron_context,
                    resource[resource_singular]['id'], reset_status)

        req = self.new_show_request(
            resource_name, resource[resource_singular]['id'], fmt=self.fmt,
            fields=fields)
        res = self.deserialize(self.fmt, req.get_response(self.ext_api))

        if not fields:
            self.assertEqual(NEW_STATUS, res[resource_singular]['status'])
            self.assertEqual(NEW_STATUS_DETAILS,
                             res[resource_singular]['status_details'])
        elif not gplugin.STATUS_SET.intersection(set(fields)):
            db_obj = getattr(
                gpmdb.GroupPolicyMappingDbPlugin, "get_" + resource_singular)(
                self._gbp_plugin, neutron_context,
                resource[resource_singular]['id'])
            self.assertIsNone(db_obj['status'])
            self.assertIsNone(db_obj['status_details'])

    def test_status_change_on_get(self):
        for resource_name in gpolicy.RESOURCE_ATTRIBUTE_MAP:
            self._test_status_change_on_get(resource_name)

    def test_no_status_change_on_get(self):
        # We explicitly populate the fields list with no status attributes
        for resource_name in gpolicy.RESOURCE_ATTRIBUTE_MAP:
            self._test_status_change_on_get(resource_name,
                                            fields=['name'])

    def _test_status_change_on_list(self, resource_name, fields=None):
        resource_singular = self._get_resource_singular(resource_name)
        if resource_name == 'policy_rules':
            pc_id = self.create_policy_classifier()['policy_classifier']['id']
            objs = [self.create_policy_rule(policy_classifier_id=pc_id),
                    self.create_policy_rule(policy_classifier_id=pc_id),
                    self.create_policy_rule(policy_classifier_id=pc_id)]
        else:
            create_method = "create_" + resource_singular
            objs = [getattr(self, create_method)(),
                    getattr(self, create_method)(),
                    getattr(self, create_method)()]

        neutron_context = context.Context('', self._tenant_id)
        reset_status = {resource_singular: {'status': None,
                                            'status_details': None}}
        for obj in objs:
            # Reset status directly in the DB to test that GET works
            getattr(gpmdb.GroupPolicyMappingDbPlugin,
                    "update_" + resource_singular)(
                        self._gbp_plugin, neutron_context,
                        obj[resource_singular]['id'], reset_status)
            req = self.new_show_request(
                resource_name, obj[resource_singular]['id'], fmt=self.fmt,
                fields=fields)
            res = self.deserialize(self.fmt, req.get_response(self.ext_api))

            if not fields:
                self.assertEqual(NEW_STATUS, res[resource_singular]['status'])
                self.assertEqual(NEW_STATUS_DETAILS,
                                 res[resource_singular]['status_details'])
            elif not gplugin.STATUS_SET.intersection(set(fields)):
                db_obj = getattr(
                    gpmdb.GroupPolicyMappingDbPlugin,
                    "get_" + resource_singular)(
                    self._gbp_plugin, neutron_context,
                    obj[resource_singular]['id'])
                self.assertIsNone(db_obj['status'])
                self.assertIsNone(db_obj['status_details'])

    def test_status_change_on_list(self):
        for resource_name in gpolicy.RESOURCE_ATTRIBUTE_MAP:
            self._test_status_change_on_list(resource_name)

    def test_no_status_change_on_list(self):
        for resource_name in gpolicy.RESOURCE_ATTRIBUTE_MAP:
            self._test_status_change_on_list(resource_name, fields=['name'])


class TestPolicyAction(GroupPolicyPluginTestCase):

    def test_redirect_value(self):
        scs_id = self._create_servicechain_spec(
            node_types=['FIREWALL_TRANSPARENT'])
        res = self.create_policy_action(action_type='redirect',
                                        action_value=scs_id, shared=True,
                                        expected_res_status=400)
        self.assertEqual(
            'SharedResourceReferenceError', res['NeutronError']['type'])

        res = self.create_policy_action(
            action_type='redirect', action_value=scs_id, tenant_id='different',
            expected_res_status=404)
        self.assertEqual(
            'ServiceChainSpecNotFound', res['NeutronError']['type'])

        res = self.create_policy_action(
            action_type='redirect', action_value=scs_id, tenant_id='different',
            expected_res_status=201, is_admin_context=True)

        res = self.create_policy_action(
            action_type='redirect', action_value=scs_id,
            expected_res_status=201)['policy_action']
        res = self.update_policy_action(
            res['id'], shared=True, expected_res_status=400)
        self.assertEqual(
            'SharedResourceReferenceError', res['NeutronError']['type'])

        scs_id = self._create_servicechain_spec(
            node_types=['FIREWALL_TRANSPARENT'], shared=True)
        self.create_policy_action(
            action_type='redirect', action_value=scs_id, shared=True,
            expected_res_status=201)
        data = {'servicechain_spec': {'shared': False}}
        scs_req = self.new_update_request(
            SERVICECHAIN_SPECS, data, scs_id, self.fmt)
        res = self.deserialize(
            self.fmt, scs_req.get_response(self.ext_api))
        self.assertEqual(
            'InvalidSharedAttributeUpdate', res['NeutronError']['type'])

    def test_redirect_shared_create(self):
        scs_id = self._create_servicechain_spec(
            node_types=['FIREWALL_TRANSPARENT'], shared=True)
        self.create_policy_action(action_type='redirect', action_value=scs_id,
                                  shared=True, expected_res_status=201)


class TestGroupPolicyPluginGroupResources(
        GroupPolicyPluginTestCase, tgpdb.TestGroupResources):

    pass


class TestGroupPolicyPluginMappedGroupResourceAttrs(
        GroupPolicyPluginTestCase, tgpmdb.TestMappedGroupResourceAttrs):

    pass


class TestQuotasForGBP(GroupPolicyPluginTestCase):

    def setUp(self, core_plugin=None, gp_plugin=None):
        cfg.CONF.set_override('quota_l3_policy', 1, group='QUOTAS')
        cfg.CONF.set_override('quota_l2_policy', 1, group='QUOTAS')
        cfg.CONF.set_override('quota_policy_target_group', 1, group='QUOTAS')
        cfg.CONF.set_override('quota_policy_target', 1, group='QUOTAS')
        cfg.CONF.set_override('quota_policy_action', 1, group='QUOTAS')
        cfg.CONF.set_override('quota_policy_classifier', 1, group='QUOTAS')
        cfg.CONF.set_override('quota_policy_rule', 1, group='QUOTAS')
        cfg.CONF.set_override('quota_policy_rule_set', 1, group='QUOTAS')
        cfg.CONF.set_override('quota_network_service_policy', 1,
                              group='QUOTAS')
        cfg.CONF.set_override('quota_external_policy', 1, group='QUOTAS')
        cfg.CONF.set_override('quota_external_segment', 1, group='QUOTAS')
        cfg.CONF.set_override('quota_nat_pool', 1, group='QUOTAS')
        super(TestQuotasForGBP, self).setUp(
            core_plugin=core_plugin, gp_plugin=gp_plugin)

    def tearDown(self):
        cfg.CONF.set_override('quota_l3_policy', -1, group='QUOTAS')
        cfg.CONF.set_override('quota_l2_policy', -1, group='QUOTAS')
        cfg.CONF.set_override('quota_policy_target_group', -1, group='QUOTAS')
        cfg.CONF.set_override('quota_policy_target', -1, group='QUOTAS')
        cfg.CONF.set_override('quota_policy_action', -1, group='QUOTAS')
        cfg.CONF.set_override('quota_policy_classifier', -1, group='QUOTAS')
        cfg.CONF.set_override('quota_policy_rule', -1, group='QUOTAS')
        cfg.CONF.set_override('quota_policy_rule_set', -1, group='QUOTAS')
        cfg.CONF.set_override('quota_network_service_policy', -1,
                              group='QUOTAS')
        cfg.CONF.set_override('quota_external_policy', -1, group='QUOTAS')
        cfg.CONF.set_override('quota_external_segment', -1, group='QUOTAS')
        cfg.CONF.set_override('quota_nat_pool', -1, group='QUOTAS')
        super(TestQuotasForGBP, self).tearDown()

    def test_group_resources_quota(self):
        ptg_id = self.create_policy_target_group()['policy_target_group']['id']
        self.create_policy_target(policy_target_group_id=ptg_id)
        self.assertRaises(webob.exc.HTTPClientError,
                          self.create_policy_target,
                          policy_target_group_id=ptg_id)
        self.assertRaises(webob.exc.HTTPClientError,
                          self.create_policy_target_group)

    def test_l3policy_quota(self):
        self.create_l3_policy()
        self.assertRaises(webob.exc.HTTPClientError,
                          self.create_l3_policy)

    def test_l2policy_quota(self):
        self.create_l2_policy()
        self.assertRaises(webob.exc.HTTPClientError,
                          self.create_l2_policy)

    def test_policy_resources_quota(self):
        pa_id = self.create_policy_action()['policy_action']['id']
        pc_id = self.create_policy_classifier()['policy_classifier']['id']
        pr_id = self.create_policy_rule(
            policy_classifier_id=pc_id,
            policy_actions=[pa_id])['policy_rule']['id']
        self.create_policy_rule_set(policy_rules=[pr_id])
        self.assertRaises(webob.exc.HTTPClientError,
                          self.create_policy_action)
        self.assertRaises(webob.exc.HTTPClientError,
                          self.create_policy_classifier)
        self.assertRaises(webob.exc.HTTPClientError,
                          self.create_policy_rule,
                          policy_classifier_id=pc_id,
                          policy_actions=[pa_id])
        self.assertRaises(webob.exc.HTTPClientError,
                          self.create_policy_rule_set,
                          policy_rules=[pr_id])
        self.create_network_service_policy()
        self.assertRaises(webob.exc.HTTPClientError,
                          self.create_network_service_policy)

    def test_external_connectivity_resources_quota(self):
        self.create_external_policy()
        self.create_external_segment()
        self.create_nat_pool()
        self.assertRaises(webob.exc.HTTPClientError,
                          self.create_external_policy)
        self.assertRaises(webob.exc.HTTPClientError,
                          self.create_external_segment)
        self.assertRaises(webob.exc.HTTPClientError,
                          self.create_nat_pool)



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
"""ncp_sci_nf_status

Revision ID: bff1774e749e
Revises: cb5b16acbeb0
Create Date: 2017-02-24 00:16:12.276236

"""

# revision identifiers, used by Alembic.
revision = 'bff1774e749e'
down_revision = 'cb5b16acbeb0'

from alembic import op
import sqlalchemy as sa


def upgrade():
    op.drop_constraint('PRIMARY',
                       'ncp_node_instance_network_function_mappings',
                       type_='primary')
    op.create_primary_key("ncp_node_instance_network_function_mappings_pk",
                          "ncp_node_instance_network_function_mappings",
                          ['sc_instance_id', 'sc_node_id'])
    op.alter_column('ncp_node_instance_network_function_mappings',
                    'network_function_id',
                    nullable=True, existing_type=sa.String(length=36))
    op.add_column('ncp_node_instance_network_function_mappings',
                  sa.Column('status', sa.String(length=50), nullable=True))
    op.add_column('ncp_node_instance_network_function_mappings',
                  sa.Column('status_details', sa.String(length=4096),
                      nullable=True))


def downgrade():
    pass



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""HA IP address to Port ID association

Revision ID: 4c0c1e2c0160
Revises: 27b724002081
Create Date: 2015-10-19 02:08:54.252877

"""

# revision identifiers, used by Alembic.
revision = '4c0c1e2c0160'
down_revision = '27b724002081'

from alembic import op
import sqlalchemy as sa

from neutron.db import migration


def upgrade():
    if not migration.schema_has_table('apic_ml2_ha_ipaddress_to_port_owner'):
        op.create_table(
            'apic_ml2_ha_ipaddress_to_port_owner',
            sa.Column('ha_ip_address', sa.String(length=64), nullable=False),
            sa.Column('port_id', sa.String(length=64), nullable=False),
            sa.ForeignKeyConstraint(
                ['port_id'], ['ports.id'], ondelete='CASCADE',
                name='apic_ml2_ha_ipaddress_to_port_owner_fk_port_id'),
            sa.PrimaryKeyConstraint('ha_ip_address', 'port_id'))



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

"""apic_aim_persist

Revision ID: cc09261e0fb5
Revises: c460c5682e74
Create Date: 2017-05-15 00:00:00.000000

"""

# revision identifiers, used by Alembic.
revision = 'cc09261e0fb5'
down_revision = 'c460c5682e74'

from alembic import op
from alembic import util
import sqlalchemy as sa


def upgrade():

    op.create_table(
        'apic_aim_address_scope_mappings',
        sa.Column('scope_id', sa.String(36), nullable=False),
        sa.Column('vrf_name', sa.String(64), nullable=True),
        sa.Column('vrf_tenant_name', sa.String(64), nullable=True),
        sa.Column('vrf_owned', sa.Boolean, nullable=False),
        sa.ForeignKeyConstraint(
            ['scope_id'], ['address_scopes.id'],
            name='apic_aim_address_scope_mappings_fk_scope_id',
            ondelete='CASCADE'),
        sa.PrimaryKeyConstraint('scope_id'))

    op.create_table(
        'apic_aim_network_mappings',
        sa.Column('network_id', sa.String(36), nullable=False),
        sa.Column('bd_name', sa.String(64), nullable=True),
        sa.Column('bd_tenant_name', sa.String(64), nullable=True),
        sa.Column('epg_name', sa.String(64), nullable=True),
        sa.Column('epg_tenant_name', sa.String(64), nullable=True),
        sa.Column('epg_app_profile_name', sa.String(64), nullable=True),
        sa.Column('vrf_name', sa.String(64), nullable=True),
        sa.Column('vrf_tenant_name', sa.String(64), nullable=True),
        sa.ForeignKeyConstraint(
            ['network_id'], ['networks.id'],
            name='apic_aim_network_mappings_fk_network_id',
            ondelete='CASCADE'),
        sa.PrimaryKeyConstraint('network_id'))

    # See if AIM is being used, and if so, migrate data.
    bind = op.get_bind()
    insp = sa.engine.reflection.Inspector.from_engine(bind)
    if 'aim_tenants' in insp.get_table_names():
        try:
            # Note - this cannot be imported unless we know the
            # apic_aim mechanism driver is deployed, since the AIM
            # library may not be installed.
            from gbpservice.neutron.plugins.ml2plus.drivers.apic_aim import (
                data_migrations)

            session = sa.orm.Session(bind=bind, autocommit=True)
            data_migrations.do_apic_aim_persist_migration(session)
        except ImportError:
            util.warning("AIM schema present, but failed to import AIM"
                        "libraries - data not migrated.")

    op.drop_table('apic_aim_addr_scope_extensions')


def downgrade():
    pass



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

"""bgp with svi

Revision ID: 804d991a3564
Revises: 5fe2285071fd
Create Date: 2018-02-06 14:31:34.104417

"""

# revision identifiers, used by Alembic.
revision = '804d991a3564'
down_revision = '5fe2285071fd'

from alembic import op
import sqlalchemy as sa


def upgrade():
    op.add_column('apic_aim_network_extensions',
                  sa.Column('bgp_enable', sa.Boolean,
                            server_default=sa.false(),
                            nullable=False))
    op.add_column('apic_aim_network_extensions',
                  sa.Column('bgp_type', sa.Enum('default_export', ''),
                            server_default="default_export", nullable=False))
    op.add_column('apic_aim_network_extensions',
                  sa.Column('bgp_asn', sa.String(64), server_default="0",
                            nullable=False))


def downgrade():
    pass



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import json

from unittest import mock

from neutron.tests import base
from oslo_serialization import jsonutils
import requests
import six

from gbpservice.contrib.nfp.configurator.agents import vpn
from gbpservice.contrib.nfp.configurator.drivers.base import base_driver
from gbpservice.contrib.nfp.configurator.drivers.vpn.vyos import (
    vyos_vpn_driver)
from gbpservice.contrib.tests.unit.nfp.configurator.test_data import (
    vpn_test_data)


bdobj = base_driver.BaseDriver('conf')
bdobj.register_agent_object_with_driver(
    'agent',
    vpn.VpnaasRpcSender(vpn_test_data.VPNTestData().sc))


class VpnaasIpsecDriverTestCase(base.BaseTestCase):
    """
    Implements test cases for driver methods
    of vpn.

    """
    def __init__(self, *args, **kwargs):
        super(VpnaasIpsecDriverTestCase, self).__init__(*args, **kwargs)
        self.conf = 'conf'
        self.test_dict = vpn_test_data.VPNTestData()
        self.context = self.test_dict.make_service_context()
        self.headers = self.test_dict.get_header()
        self.plugin_rpc = vpn.VpnaasRpcSender(self.test_dict.sc)
        self.driver = vyos_vpn_driver.VpnaasIpsecDriver(self.conf)
        self.svc_validate = (
            vyos_vpn_driver.VPNServiceValidator(self.plugin_rpc))
        self.resp = mock.Mock()
        self.fake_resp_dict = {'status': True}

    def test_create_vpn_service(self):
        """
        Implements method to test the vpn driver's create vpn service.
        """

        context = self.test_dict.make_service_context(operation_type='vpn')

        kwargs = self.test_dict.make_resource_data(operation='create',
                                                   service_type='vpn')
        with mock.patch.object(bdobj.agent, 'update_status') as (
                mock_update_status):
            self.driver.vpnservice_updated(context, kwargs)
            mock_update_status.assert_called_with(
                self.context,
                self.test_dict.vpn_vpnsvc_active)

    def test_create_ipsec_site_conn(self):
        """
        Implements method to test the vpn driver's create ipsec site conn
        """
        self.resp = mock.Mock(status_code=200)
        context = self.test_dict.make_service_context()
        kwargs = self.test_dict.make_resource_data(operation='create',
                                                   service_type='ipsec')
        with mock.patch.object(
                bdobj.agent, 'update_status') as mock_update_status, (
            mock.patch.object(jsonutils, 'loads')) as mock_resp, (
            mock.patch.object(requests, 'post')) as mock_post, (
            mock.patch.object(
                self.driver.agent, 'get_vpn_servicecontext',
                return_value=[self.test_dict.svc_context])):
            mock_resp.return_value = self.fake_resp_dict
            mock_post.return_value = self.resp
            self.driver.vpnservice_updated(context, kwargs)

            mock_post.assert_called_with(
                self.test_dict.url_create_ipsec_conn,
                data=jsonutils.dumps(self.test_dict.svc_context),
                headers=self.headers,
                timeout=self.test_dict.timeout)
            mock_update_status.assert_called_with(
                context,
                self.test_dict.ipsec_vpnsvc_status)

    def _dict_to_query_str(self, args):
        return '&'.join([str(k) + '=' + str(v)
                         for k, v in six.iteritems(args)])

    def test_delete_ipsec_site_conn(self):
        """
        Implements method to test the vpn driver's create ipsec site conn
        """

        self.resp = mock.Mock(status_code=200)
        kwargs = self.test_dict.make_resource_data(operation='delete',
                                                   service_type='ipsec')
        with mock.patch.object(self.plugin_rpc, 'ipsec_site_conn_deleted'), (
                mock.patch.object(json, 'loads')) as mock_resp, (
                mock.patch.object(requests, 'delete')) as (
                mock_delete):
            mock_resp.return_value = self.fake_resp_dict
            mock_delete.return_value = self.resp
            self.driver.vpnservice_updated(self.context, kwargs)
            resource = kwargs['resource']
            svc_desc = resource['description']
            tokens = svc_desc.split(';')
            cidr = tokens[1].split('=')[1]

            tunnel = {}
            tunnel['peer_address'] = resource['peer_address']
            tunnel['local_cidr'] = cidr
            tunnel['peer_cidrs'] = resource['peer_cidrs']

            url = self.test_dict.url_delete_ipsec_conn

            mock_delete.assert_called_with(
                url.encode('ascii', 'ignore'),
                timeout=self.test_dict.timeout,
                headers=self.headers,
                data=None)

    def test_check_status(self):
        """
        Implements method to test the vpn driver's check status
        """

        self.resp = mock.Mock(status_code=200)
        svc_context = self.test_dict.svc_context
        with mock.patch.object(self.plugin_rpc, 'update_status'), (
                mock.patch.object(self.resp, 'json')) as mock_json, (
                mock.patch.object(requests, 'get')) as mock_get:
            mock_get.return_value = self.resp
            mock_json.return_value = {'state': 'DOWN'}
            state = self.driver.check_status(self.context, svc_context)
            self.assertIsNone(state)


class VpnGenericConfigDriverTestCase(base.BaseTestCase):
    """
    Implements test cases for driver methods
    of generic config.

    """
    def __init__(self, *args, **kwargs):
        super(VpnGenericConfigDriverTestCase, self).__init__(*args, **kwargs)
        self.conf = 'conf'
        self.test_dict = vpn_test_data.VPNTestData()
        self.context = self.test_dict.make_service_context()
        self.headers = self.test_dict.get_header()
        self.plugin_rpc = vpn.VpnaasRpcSender(self.test_dict.sc)
        self.rest_apt = vyos_vpn_driver.RestApi(self.test_dict.vm_mgmt_ip)
        self.driver = vyos_vpn_driver.VpnGenericConfigDriver()
        self.resp = mock.Mock()
        self.fake_resp_dict = {'status': True}
        self.kwargs = self.test_dict.fake_resource_data()

    def setUp(self):
        super(VpnGenericConfigDriverTestCase, self).setUp()
        self.resp = mock.Mock(status_code=200)

    def tearDown(self):
        super(VpnGenericConfigDriverTestCase, self).tearDown()
        self.resp = mock.Mock(status_code=200)

    def test_configure_interfaces(self):
        """
        Implements test case for configure interfaces method
        of generic config driver.

        Returns: none

        """

        with mock.patch.object(
                requests, 'post', return_value=self.resp) as mock_post, (
            mock.patch.object(self.resp,
                              'json',
                              return_value=self.fake_resp_dict)):
            self.driver.configure_interfaces(self.test_dict.context_device,
                                             self.kwargs)

            mock_post.assert_called_with(
                self.test_dict.url_for_add_inte,
                jsonutils.dumps(
                    self.test_dict.data_for_interface),
                headers=self.headers,
                timeout=self.test_dict.timeout)

    def test_clear_interfaces(self):
        """
        Implements test case for clear interfaces method
        of generic config driver.

        Returns: none

        """

        self.resp = mock.Mock(status_code=200)
        with mock.patch.object(
                requests, 'delete', return_value=self.resp) as mock_delete, (
            mock.patch.object(
                self.resp, 'json', return_value=self.fake_resp_dict)):
            self.driver.clear_interfaces(self.test_dict.context_device,
                                         self.kwargs)

            mock_delete.assert_called_with(
                self.test_dict.url_for_del_inte,
                data=jsonutils.dumps(
                    self.test_dict.data_for_interface),
                headers=self.headers,
                timeout=self.test_dict.timeout)

    def test_configure_source_routes(self):
        """
        Implements test case for configure routes method
        of generic config driver.

        Returns: none

        """

        with mock.patch.object(
                requests, 'post', return_value=self.resp) as mock_post, (
            mock.patch.object(jsonutils, 'loads',
                              return_value=self.fake_resp_dict)):
            self.driver.configure_routes(self.test_dict.context_device,
                                         self.kwargs)

            mock_post.assert_called_with(
                self.test_dict.url_for_add_src_route,
                data=jsonutils.dumps(
                    self.test_dict.data_for_add_src_route),
                headers=self.headers,
                timeout=self.test_dict.timeout)

    def test_delete_source_routes(self):
        """
        Implements test case for clear routes method
        of generic config driver.

        Returns: none

        """

        with mock.patch.object(requests, 'post', return_value=self.resp), (
            mock.patch.object(
                requests, 'delete', return_value=self.resp)) as mock_delete:
            self.driver.clear_routes(
                self.test_dict.context_device, self.kwargs)

            mock_delete.assert_called_with(
                self.test_dict.url_for_del_src_route,
                data=jsonutils.dumps(
                    self.test_dict.data_for_del_src_route),
                headers=self.headers,
                timeout=self.test_dict.timeout)


class VPNSvcValidatorTestCase(base.BaseTestCase):

    def __init__(self, *args, **kwargs):
        super(VPNSvcValidatorTestCase, self).__init__(*args, **kwargs)
        self.test_dict = vpn_test_data.VPNTestData()
        self.plugin_rpc = vpn.VpnaasRpcSender(self.test_dict.sc)
        self.valid_obj = vyos_vpn_driver.VPNServiceValidator(self.plugin_rpc)

    def test_validate_active(self):
        """
        Implements testcase for vpn driver's validate method to test in
        success condition while making call to the service VM
        """

        context = self.test_dict.make_service_context()
        svc = self.test_dict._create_vpnservice_obj()
        description = str(svc['resource']['description'])
        description = description.split(';')
        description[1] = 'tunnel_local_cidr=12.0.6.0/24'
        description = ";".join(description)
        svc.update({'description': description})

        with mock.patch.object(self.plugin_rpc, "update_status") as mock_valid:
            self.valid_obj.validate(context, svc)
            mock_valid.assert_called_with(context,
                                          self.test_dict.vpn_vpnsvc_active)

    def test_validate_error(self):
        """
        Implements testcase for vpn driver's validate method to test in
        fail condition while making call to the service VM
        """

        context = self.test_dict.make_service_context()
        with mock.patch.object(self.plugin_rpc, "update_status") as mock_valid:
            self.valid_obj.validate(
                context,
                self.test_dict._create_vpnservice_obj())
            mock_valid.assert_called_with(
                context,
                self.test_dict.vpn_vpnsvc_active)


class RestApiTestCase(base.BaseTestCase):
    """
    Class which implements the testcases to test the vpn RestApi calls.
    """

    def __init__(self, *args, **kwargs):
        super(RestApiTestCase, self).__init__(*args, **kwargs)
        self.rest_obj = vyos_vpn_driver.RestApi((
            vpn_test_data.VPNTestData().vm_mgmt_ip))
        self.resp = mock.Mock()
        self.resp = mock.Mock(status_code=200)
        self.test_dict = vpn_test_data.VPNTestData()
        self.headers = self.test_dict.get_header()
        self.args = {'peer_address': '1.103.2.2'}
        self.fake_resp_dict = {'status': None}
        self.timeout = self.rest_obj.timeout
        self.data = {'data': 'data'}
        self.j_data = jsonutils.dumps(self.data)

    def test_post_success(self):
        """
        Implements testcase for vpn drivers post method to test in
        success condition while making call to the service VM
        """

        self.resp = mock.Mock(status_code=200)
        self.fake_resp_dict.update({'status': True})
        with mock.patch.object(requests, 'post', return_value=self.resp) as (
            mock_post), (
            mock.patch.object(jsonutils, 'loads',
                              return_value=self.fake_resp_dict)):
            self.rest_obj.post('create-ipsec-site-conn',
                               self.data, self.headers)
            mock_post.assert_called_with(
                self.test_dict.url_create_ipsec_conn,
                data=self.j_data,
                headers=self.headers,
                timeout=self.timeout)

    def test_put_success(self):
        """
        Implements testcase for vpn drivers put method to test in
        success condition while making call to the service VM
        """

        self.resp = mock.Mock(status_code=200)
        with mock.patch.object(requests, 'put', return_value=self.resp) as (
                mock_put):
            self.rest_obj.put('create-ipsec-site-conn',
                              self.data, self.headers)
            mock_put.assert_called_with(
                self.test_dict.url_create_ipsec_conn,
                data=self.j_data,
                headers=self.headers,
                timeout=self.timeout)

    def test_put_fail(self):
        """
        Implements testcase for vpn drivers put method to test in
        fail condition while making call to the service VM
        """

        self.resp = mock.Mock(status_code=404)
        with mock.patch.object(requests, 'put', return_value=self.resp) as (
                mock_put):

            self.rest_obj.put('create-ipsec-site-conn',
                              self.data, self.headers)
            mock_put.assert_called_with(
                self.test_dict.url_create_ipsec_conn,
                data=jsonutils.dumps(self.data),
                headers=self.headers,
                timeout=self.timeout)

    def test_delete_success(self):
        """
        Implements testcase for vpn drivers delete method to test in
        success condition while making call to the service VM
        """
        self.resp = mock.Mock(status_code=200)
        self.fake_resp_dict.update({'status': True})
        with mock.patch.object(requests, 'delete', return_value=self.resp) as (
            mock_delete), (
            mock.patch.object(jsonutils, 'loads',
                              return_value=self.fake_resp_dict)):
            self.rest_obj.delete('delete-ipsec-site-conn',
                                 self.args,
                                 self.headers)
            mock_delete.assert_called_with(
                self.test_dict.url_delete_ipsec_conn,
                timeout=self.timeout,
                data=None,
                headers=self.headers)

    def test_get_success(self):
        """
        Implements testcase for vpn drivers get methode to test in
        fail condition while making call to the service VM
        """

        self.resp = mock.Mock(status_code=200)
        with mock.patch.object(requests, 'get', return_value=self.resp) as (
                mock_get):
            self.rest_obj.get('create-ipsec-site-tunnel',
                              self.data, self.headers)
            mock_get.assert_called_with(
                self.test_dict.url_create_ipsec_tunnel,
                params=self.data,
                headers=self.headers,
                timeout=self.timeout)

    def test_get_fail(self):
        """
        Implements testcase for vpn drivers get methode to test in
        fail condition
        """

        self.resp = mock.Mock(status_code=404)
        with mock.patch.object(requests, 'get', return_value=self.resp) as (
                mock_get):
            self.rest_obj.get('create-ipsec-site-tunnel',
                              self.data, self.headers)
            mock_get.assert_called_with(
                self.test_dict.url_create_ipsec_tunnel,
                params=self.data,
                headers=self.headers,
                timeout=self.timeout)



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

"""alter project_id column length in gp_apic_tenant_specific_nat_epg

Revision ID: cb5b16acbeb0
Revises: da6a25bbcfa8
Create Date: 2017-06-10 00:00:00.000000

"""

# revision identifiers, used by Alembic.
revision = 'cb5b16acbeb0'
down_revision = 'da6a25bbcfa8'

from alembic import op
import sqlalchemy as sa


def upgrade():
    op.alter_column('gp_apic_tenant_specific_nat_epg',
                    'project_id',
                    existing_type=sa.String(36),
                    type_=sa.String(255), nullable=False)


def downgrade():
    pass



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import six
import webob.exc

from neutron_lib import context
from neutron_lib.plugins import constants
from oslo_config import cfg
from oslo_utils import uuidutils

from gbpservice.neutron.db import servicechain_db as svcchain_db
from gbpservice.neutron.extensions import servicechain as service_chain
from gbpservice.neutron.tests.unit import common as cm
from gbpservice.neutron.tests.unit.db.grouppolicy import test_group_policy_db

JSON_FORMAT = 'json'
GP_PLUGIN_KLASS = (
    "gbpservice.neutron.services.grouppolicy.plugin.GroupPolicyPlugin")


class ServiceChainDBTestBase(test_group_policy_db.GroupPolicyDBTestBase):

    def _get_resource_plural(self, resource):
        if resource.endswith('y'):
            resource_plural = resource.replace('y', 'ies')
        else:
            resource_plural = resource + 's'

        return resource_plural

    def _test_list_resources(self, resource, items,
                             neutron_context=None,
                             query_params=None):
        resource_plural = self._get_resource_plural(resource)

        res = self._list(resource_plural,
                         neutron_context=neutron_context,
                         query_params=query_params)
        params = query_params.split('&')
        params = dict((x.split('=')[0], x.split('=')[1].split(','))
                      for x in params)
        count = getattr(self.plugin, 'get_%s_count' % resource_plural)(
            neutron_context or context.get_admin_context(), params)
        self.assertEqual(len(res[resource_plural]), count)
        resource = resource.replace('-', '_')
        self.assertEqual(sorted([i['id'] for i in res[resource_plural]]),
                         sorted([i[resource]['id'] for i in items]))

    def _create_profiled_servicechain_node(
            self, service_type=constants.LOADBALANCERV2, shared_profile=False,
            profile_tenant_id=None, **kwargs):
        prof = self.create_service_profile(
            service_type=service_type,
            shared=shared_profile,
            tenant_id=profile_tenant_id or self._tenant_id)['service_profile']
        return self.create_servicechain_node(
            service_profile_id=prof['id'], **kwargs)


class ServiceChainDBTestPlugin(svcchain_db.ServiceChainDbPlugin):

    supported_extension_aliases = ['servicechain'] + (
        test_group_policy_db.UNSUPPORTED_REQUIRED_EXTS)
    path_prefix = "/servicechain"


DB_GP_PLUGIN_KLASS = (ServiceChainDBTestPlugin.__module__ + '.' +
                      ServiceChainDBTestPlugin.__name__)


class ServiceChainDbTestCase(test_group_policy_db.GroupPolicyDbTestCase):

    def setUp(self, core_plugin=None, sc_plugin=None, service_plugins=None,
              ext_mgr=None, gp_plugin=None):

        super(ServiceChainDbTestCase, self).setUp(
            gp_plugin=gp_plugin or GP_PLUGIN_KLASS, core_plugin=core_plugin,
            sc_plugin=sc_plugin, service_plugins=service_plugins,
            ext_mgr=ext_mgr)
        self.plugin = self._sc_plugin


class TestServiceChainResources(ServiceChainDbTestCase):

    def _test_show_resource(self, resource, resource_id, attrs):
        resource_plural = self._get_resource_plural(resource)
        req = self.new_show_request(resource_plural, resource_id,
                                    fmt=self.fmt)
        res = self.deserialize(self.fmt,
                               req.get_response(self.ext_api))

        for k, v in six.iteritems(attrs):
            self.assertEqual(v, res[resource][k])

    def test_create_servicechain_specs_same_node(self):
        template1 = '{"key1":"value1"}'
        sp = self.create_service_profile(
            service_type=constants.FIREWALL)['service_profile']
        scn = self.create_servicechain_node(
            config=template1, service_profile_id=sp['id'])
        scn_id = scn['servicechain_node']['id']
        spec1 = {"servicechain_spec": {'name': 'scs1',
                                       'tenant_id': self._tenant_id,
                                       'nodes': [scn_id]}}
        spec_req = self.new_create_request('servicechain_specs',
                                           spec1,
                                           self.fmt)
        spec_res = spec_req.get_response(self.ext_api)
        self.assertEqual(webob.exc.HTTPCreated.code, spec_res.status_int)
        res = self.deserialize(self.fmt, spec_res)
        self.assertIn('servicechain_spec', res)
        self.assertEqual([scn_id], res['servicechain_spec']['nodes'])
        spec2 = {"servicechain_spec": {'name': 'scs2',
                                       'tenant_id': self._tenant_id,
                                       'nodes': [scn_id]}}
        spec_req = self.new_create_request('servicechain_specs',
                                           spec2,
                                           self.fmt)
        spec_res = spec_req.get_response(self.ext_api)
        self.assertEqual(webob.exc.HTTPCreated.code, spec_res.status_int)
        res = self.deserialize(self.fmt, spec_res)
        self.assertIn('servicechain_spec', res)
        self.assertEqual([scn_id], res['servicechain_spec']['nodes'])

    def test_create_and_show_servicechain_node(self):
        profile = self.create_service_profile(service_type=constants.FIREWALL)
        attrs = cm.get_create_servicechain_node_default_attrs(
            service_profile_id=profile['service_profile']['id'],
            config="config1")

        scn = self.create_servicechain_node(
            service_profile_id=profile['service_profile']['id'],
            config="config1")

        for k, v in six.iteritems(attrs):
            self.assertEqual(v, scn['servicechain_node'][k])

        self._test_show_resource('servicechain_node',
                                 scn['servicechain_node']['id'],
                                 attrs)

    def test_list_servicechain_nodes(self):
        scns = [
            self._create_profiled_servicechain_node(name='scn1',
                                                    description='scn'),
            self._create_profiled_servicechain_node(name='scn2',
                                                    description='scn'),
            self._create_profiled_servicechain_node(name='scn3',
                                                    description='scn')]
        self._test_list_resources('servicechain_node', scns,
                                  query_params='description=scn')

    def test_update_servicechain_node(self):
        name = 'new_servicechain_node'
        description = 'new desc'
        config = 'new_config'
        profile = self.create_service_profile(service_type=constants.FIREWALL)
        attrs = cm.get_create_servicechain_node_default_attrs(
            name=name, description=description,
            config=config,
            service_profile_id=profile['service_profile']['id'])

        scn = self.create_servicechain_node(
            service_profile_id=profile['service_profile']['id'])

        data = {'servicechain_node': {'name': name,
                                      'description': description,
                                      'config': config}}
        req = self.new_update_request('servicechain_nodes', data,
                                      scn['servicechain_node']['id'])
        res = self.deserialize(self.fmt, req.get_response(self.ext_api))

        for k, v in six.iteritems(attrs):
            self.assertEqual(v, res['servicechain_node'][k])

        self._test_show_resource('servicechain_node',
                                 scn['servicechain_node']['id'],
                                 attrs)

    def test_delete_servicechain_node(self):
        ctx = context.get_admin_context()

        scn = self._create_profiled_servicechain_node()
        scn_id = scn['servicechain_node']['id']

        scs = self.create_servicechain_spec(nodes=[scn_id])
        scs_id = scs['servicechain_spec']['id']

        # Deleting Service Chain Node in use by a Spec should fail
        self.assertRaises(service_chain.ServiceChainNodeInUse,
                          self.plugin.delete_servicechain_node, ctx, scn_id)

        req = self.new_delete_request('servicechain_specs', scs_id)
        res = req.get_response(self.ext_api)
        self.assertEqual(webob.exc.HTTPNoContent.code, res.status_int)

        # After deleting the Service Chain Spec, node delete should succeed
        req = self.new_delete_request('servicechain_nodes', scn_id)
        res = req.get_response(self.ext_api)
        self.assertEqual(webob.exc.HTTPNoContent.code, res.status_int)
        self.assertRaises(service_chain.ServiceChainNodeNotFound,
                          self.plugin.get_servicechain_node,
                          ctx, scn_id)

    def test_create_and_show_servicechain_spec(self):
        name = "scs1"
        scn = self._create_profiled_servicechain_node()
        scn_id = scn['servicechain_node']['id']

        attrs = cm.get_create_servicechain_spec_default_attrs(
            name=name, nodes=[scn_id])

        scs = self.create_servicechain_spec(name=name, nodes=[scn_id])

        for k, v in six.iteritems(attrs):
            self.assertEqual(v, scs['servicechain_spec'][k])

        self._test_show_resource('servicechain_spec',
                                 scs['servicechain_spec']['id'],
                                 attrs)

    def test_create_spec_multiple_nodes(self):
        name = "scs1"
        scn1 = self._create_profiled_servicechain_node()
        scn1_id = scn1['servicechain_node']['id']
        scn2 = self._create_profiled_servicechain_node()
        scn2_id = scn2['servicechain_node']['id']
        attrs = cm.get_create_servicechain_spec_default_attrs(
            name=name, nodes=[scn1_id, scn2_id])
        scs = self.create_servicechain_spec(
            name=name, nodes=[scn1_id, scn2_id])
        for k, v in six.iteritems(attrs):
            self.assertEqual(v, scs['servicechain_spec'][k])

    def test_list_servicechain_specs(self):
        scs = [self.create_servicechain_spec(name='scs1', description='scs'),
               self.create_servicechain_spec(name='scs2', description='scs'),
               self.create_servicechain_spec(name='scs3', description='scs')]
        self._test_list_resources('servicechain_spec', scs,
                                  query_params='description=scs')

    def test_node_ordering_list_servicechain_specs(self):
        scn1_id = self._create_profiled_servicechain_node()[
            'servicechain_node']['id']
        scn2_id = self._create_profiled_servicechain_node()[
            'servicechain_node']['id']
        nodes_list = [scn1_id, scn2_id]
        scs = self.create_servicechain_spec(name='scs1',
                                            nodes=nodes_list)
        self.assertEqual(nodes_list, scs['servicechain_spec']['nodes'])
        res = self._list('servicechain_specs')
        self.assertEqual(1, len(res['servicechain_specs']))
        self.assertEqual(nodes_list, res['servicechain_specs'][0]['nodes'])

        # Delete the service chain spec and create another with nodes in
        # reverse order and verify that that proper ordering is maintained
        req = self.new_delete_request('servicechain_specs',
                                      scs['servicechain_spec']['id'])
        res = req.get_response(self.ext_api)
        self.assertEqual(webob.exc.HTTPNoContent.code, res.status_int)

        nodes_list.reverse()
        scs = self.create_servicechain_spec(name='scs1',
                                            nodes=nodes_list)
        self.assertEqual(scs['servicechain_spec']['nodes'], nodes_list)
        res = self._list('servicechain_specs')
        self.assertEqual(1, len(res['servicechain_specs']))
        self.assertEqual(nodes_list, res['servicechain_specs'][0]['nodes'])

    def test_update_servicechain_spec(self):
        name = "new_servicechain_spec1"
        description = 'new desc'
        scn_id = self._create_profiled_servicechain_node()[
            'servicechain_node']['id']
        attrs = cm.get_create_servicechain_spec_default_attrs(
            name=name, description=description, nodes=[scn_id])
        scs = self.create_servicechain_spec()
        data = {'servicechain_spec': {'name': name, 'description': description,
                                      'nodes': [scn_id]}}
        req = self.new_update_request('servicechain_specs', data,
                                      scs['servicechain_spec']['id'])
        res = self.deserialize(self.fmt, req.get_response(self.ext_api))

        for k, v in six.iteritems(attrs):
            self.assertEqual(v, res['servicechain_spec'][k])

        self._test_show_resource('servicechain_spec',
                                 scs['servicechain_spec']['id'], attrs)

    def test_delete_servicechain_spec(self):
        ctx = context.get_admin_context()

        scs = self.create_servicechain_spec()
        scs_id = scs['servicechain_spec']['id']

        req = self.new_delete_request('servicechain_specs', scs_id)
        res = req.get_response(self.ext_api)
        self.assertEqual(webob.exc.HTTPNoContent.code, res.status_int)
        self.assertRaises(service_chain.ServiceChainSpecNotFound,
                          self.plugin.get_servicechain_spec, ctx, scs_id)

    def test_delete_spec_in_use_by_policy_action_rejected(self):
        ctx = context.get_admin_context()
        scs_id = self.create_servicechain_spec()['servicechain_spec']['id']
        data = {'policy_action': {'action_type': 'redirect',
                                  'tenant_id': self._tenant_id,
                                  'action_value': scs_id}}
        pa_req = self.new_create_request('grouppolicy/policy_actions',
                                         data, self.fmt)
        res = pa_req.get_response(self.ext_api)
        if res.status_int >= webob.exc.HTTPClientError.code:
            raise webob.exc.HTTPClientError(code=res.status_int)

        self.assertRaises(service_chain.ServiceChainSpecInUse,
                          self.plugin.delete_servicechain_spec, ctx, scs_id)

    def test_delete_spec_in_use_by_instance_rejected(self):
        ctx = context.get_admin_context()
        scs_id = self.create_servicechain_spec()['servicechain_spec']['id']

        sci = self.create_servicechain_instance(servicechain_specs=[scs_id])
        sci_id = sci['servicechain_instance']['id']

        # Deleting the Spec used by Instance should not be allowed
        self.assertRaises(service_chain.ServiceChainSpecInUse,
                          self.plugin.delete_servicechain_spec, ctx, scs_id)

        req = self.new_delete_request('servicechain_instances', sci_id)
        res = req.get_response(self.ext_api)
        self.assertEqual(webob.exc.HTTPNoContent.code, res.status_int)
        self.assertRaises(service_chain.ServiceChainInstanceNotFound,
                          self.plugin.get_servicechain_instance,
                          ctx, sci_id)

        # Deleting the spec should succeed after the instance is deleted
        req = self.new_delete_request('servicechain_specs', scs_id)
        res = req.get_response(self.ext_api)
        self.assertEqual(webob.exc.HTTPNoContent.code, res.status_int)
        self.assertRaises(service_chain.ServiceChainSpecNotFound,
                          self.plugin.get_servicechain_spec, ctx, scs_id)

    def test_create_and_show_servicechain_instance(self):
        scs_id = self.create_servicechain_spec()['servicechain_spec']['id']
        policy_target_group_id = uuidutils.generate_uuid()
        classifier_id = uuidutils.generate_uuid()
        config_param_values = "{}"
        attrs = cm.get_create_servicechain_instance_default_attrs(
            servicechain_specs=[scs_id],
            provider_ptg_id=policy_target_group_id,
            consumer_ptg_id=policy_target_group_id,
            management_ptg_id=policy_target_group_id,
            classifier_id=classifier_id,
            config_param_values=config_param_values)

        sci = self.create_servicechain_instance(
            servicechain_specs=[scs_id],
            provider_ptg_id=policy_target_group_id,
            consumer_ptg_id=policy_target_group_id,
            management_ptg_id=policy_target_group_id,
            classifier_id=classifier_id,
            config_param_values=config_param_values)
        for k, v in six.iteritems(attrs):
            self.assertEqual(v, sci['servicechain_instance'][k])

        self._test_show_resource('servicechain_instance',
                                 sci['servicechain_instance']['id'],
                                 attrs)
        req = self.new_delete_request('servicechain_instances',
                                      sci['servicechain_instance']['id'])
        req.get_response(self.ext_api)

    def test_list_servicechain_instances(self):
        servicechain_instances = [self.create_servicechain_instance(
            name='sci1', description='sci'),
            self.create_servicechain_instance(name='sci2', description='sci'),
            self.create_servicechain_instance(name='sci3', description='sci')]
        self._test_list_resources('servicechain_instance',
                                  servicechain_instances,
                                  query_params='description=sci')

    def test_spec_ordering_list_servicechain_instances(self):
        scs1_id = self.create_servicechain_spec()['servicechain_spec']['id']
        scs2_id = self.create_servicechain_spec()['servicechain_spec']['id']
        specs_list = [scs1_id, scs2_id]
        sci = self.create_servicechain_instance(name='sci1',
                                                servicechain_specs=specs_list)
        self.assertEqual(specs_list,
                         sci['servicechain_instance']['servicechain_specs'])
        res = self._list('servicechain_instances')
        self.assertEqual(1, len(res['servicechain_instances']))
        result_instance = res['servicechain_instances'][0]
        self.assertEqual(specs_list, result_instance['servicechain_specs'])

        # Delete the service chain instance and create another with specs in
        # reverse order and verify that that proper ordering is maintained
        req = self.new_delete_request('servicechain_instances',
                                      sci['servicechain_instance']['id'])
        res = req.get_response(self.ext_api)
        self.assertEqual(webob.exc.HTTPNoContent.code, res.status_int)

        specs_list.reverse()
        sci = self.create_servicechain_instance(name='sci1',
                                                servicechain_specs=specs_list)
        self.assertEqual(specs_list,
                         sci['servicechain_instance']['servicechain_specs'])
        res = self._list('servicechain_instances')
        self.assertEqual(1, len(res['servicechain_instances']))
        result_instance = res['servicechain_instances'][0]
        self.assertEqual(specs_list,
                         result_instance['servicechain_specs'])

    def test_update_servicechain_instance(self):
        name = "new_servicechain_instance"
        description = 'new desc'
        config_param_values = "{}"
        scs_id = self.create_servicechain_spec()['servicechain_spec']['id']
        provider_ptg_id = uuidutils.generate_uuid()
        consumer_ptg_id = uuidutils.generate_uuid()
        management_ptg_id = uuidutils.generate_uuid()
        classifier_id = uuidutils.generate_uuid()
        attrs = cm.get_create_servicechain_instance_default_attrs(
            name=name, description=description, servicechain_specs=[scs_id],
            provider_ptg_id=provider_ptg_id, consumer_ptg_id=consumer_ptg_id,
            management_ptg_id=management_ptg_id,
            classifier_id=classifier_id,
            config_param_values=config_param_values)

        sci = self.create_servicechain_instance(
            servicechain_specs=[scs_id], provider_ptg_id=provider_ptg_id,
            consumer_ptg_id=consumer_ptg_id,
            management_ptg_id=management_ptg_id, classifier_id=classifier_id,
            config_param_values=config_param_values)
        new_classifier_id = uuidutils.generate_uuid()
        new_scs_id = self.create_servicechain_spec()['servicechain_spec']['id']
        attrs.update({'servicechain_specs': [new_scs_id],
                      'classifier_id': new_classifier_id})
        data = {'servicechain_instance': {'name': name,
                                          'description': description,
                                          'servicechain_specs': [new_scs_id],
                                          'classifier_id': new_classifier_id}}
        req = self.new_update_request('servicechain_instances', data,
                                      sci['servicechain_instance']['id'])
        res = self.deserialize(self.fmt, req.get_response(self.ext_api))
        for k, v in six.iteritems(attrs):
            self.assertEqual(v, res['servicechain_instance'][k])

        self._test_show_resource('servicechain_instance',
                                 sci['servicechain_instance']['id'], attrs)
        req = self.new_delete_request('servicechain_instances',
                                      sci['servicechain_instance']['id'])
        req.get_response(self.ext_api)

    def test_delete_servicechain_instance(self):
        ctx = context.get_admin_context()

        sci = self.create_servicechain_instance()
        sci_id = sci['servicechain_instance']['id']

        req = self.new_delete_request('servicechain_instances', sci_id)
        res = req.get_response(self.ext_api)
        self.assertEqual(webob.exc.HTTPNoContent.code, res.status_int)
        self.assertRaises(service_chain.ServiceChainInstanceNotFound,
                          self.plugin.get_servicechain_instance,
                          ctx, sci_id)

    def test_create_and_show_service_profile(self):
        attrs = cm.get_create_service_profile_default_attrs(
            service_type=constants.FIREWALL, vendor="vendor1")

        scn = self.create_service_profile(
            service_type=constants.FIREWALL, vendor="vendor1")

        for k, v in six.iteritems(attrs):
            self.assertEqual(scn['service_profile'][k], v)

        self._test_show_resource('service_profile',
                                 scn['service_profile']['id'], attrs)

    def test_list_service_profile(self):
        scns = [self.create_service_profile(name='sp1', description='sp',
                                            service_type='LOADBALANCERV2'),
                self.create_service_profile(name='sp2', description='sp',
                                            service_type='LOADBALANCERV2'),
                self.create_service_profile(name='sp3', description='sp',
                                            service_type='LOADBALANCERV2')]
        self._test_list_resources('service_profile', scns,
                                  query_params='description=sp')

    def test_update_service_profile(self):
        name = 'new_service_profile'
        description = 'new desc'
        attrs = cm.get_create_service_profile_default_attrs(
            name=name, description=description,
            service_type=constants.FIREWALL)

        scn = self.create_service_profile(service_type=constants.FIREWALL)

        data = {'service_profile': {'name': name,
                                    'description': description}}
        req = self.new_update_request('service_profiles', data,
                                      scn['service_profile']['id'])
        res = self.deserialize(self.fmt, req.get_response(self.ext_api))

        for k, v in six.iteritems(attrs):
            self.assertEqual(res['service_profile'][k], v)

        self._test_show_resource('service_profile',
                                 scn['service_profile']['id'], attrs)

    def test_delete_service_profile(self):
        ctx = context.get_admin_context()

        sp = self.create_service_profile(service_type='LOADBALANCERV2')
        sp_id = sp['service_profile']['id']

        scn = self.create_servicechain_node(service_profile_id=sp_id)
        scn_id = scn['servicechain_node']['id']

        # Deleting Service Chain Node in use by a Spec should fail
        self.assertRaises(service_chain.ServiceProfileInUse,
                          self.plugin.delete_service_profile, ctx, sp_id)

        req = self.new_delete_request('servicechain_nodes', scn_id)
        res = req.get_response(self.ext_api)
        self.assertEqual(webob.exc.HTTPNoContent.code, res.status_int)

        # After deleting the Service Chain Spec, node delete should succeed
        req = self.new_delete_request('service_profiles', sp_id)
        res = req.get_response(self.ext_api)
        self.assertEqual(webob.exc.HTTPNoContent.code, res.status_int)
        self.assertRaises(service_chain.ServiceProfileNotFound,
                          self.plugin.get_service_profile,
                          ctx, sp_id)


class TestServiceChainStatusAttributesForResources(
    test_group_policy_db.TestStatusAttributesForResources):

    def test_set_status_attrs(self):
        for resource_name in service_chain.RESOURCE_ATTRIBUTE_MAP:
            self._test_set_status_attrs(self._get_resource_singular(
                resource_name), self._sc_plugin)


class TestQuotasForServiceChain(ServiceChainDbTestCase):

    def setUp(self, core_plugin=None, sc_plugin=None,
              gp_plugin=None, service_plugins=None, ext_mgr=None):
        cfg.CONF.set_override('quota_servicechain_node', 1,
                              group='QUOTAS')
        cfg.CONF.set_override('quota_servicechain_spec', 1,
                              group='QUOTAS')
        cfg.CONF.set_override('quota_servicechain_instance', 1,
                              group='QUOTAS')
        cfg.CONF.set_override('quota_service_profile', 1,
                              group='QUOTAS')
        super(TestQuotasForServiceChain, self).setUp(
            core_plugin=core_plugin, sc_plugin=sc_plugin,
            gp_plugin=gp_plugin, service_plugins=service_plugins,
            ext_mgr=ext_mgr)

    def tearDown(self):
        cfg.CONF.set_override('quota_servicechain_node', -1,
                              group='QUOTAS')
        cfg.CONF.set_override('quota_servicechain_spec', -1,
                              group='QUOTAS')
        cfg.CONF.set_override('quota_servicechain_instance', -1,
                              group='QUOTAS')
        cfg.CONF.set_override('quota_service_profile', -1,
                              group='QUOTAS')
        super(TestQuotasForServiceChain, self).tearDown()

    def test_servicechain_node_quota(self):
        self.create_servicechain_node()
        self.assertRaises(webob.exc.HTTPClientError,
                          self.create_servicechain_node)

    def test_servicechain_spec_quota(self):
        self.create_servicechain_spec()
        self.assertRaises(webob.exc.HTTPClientError,
                          self.create_servicechain_spec)

    def test_servicechain_instance_quota(self):
        self.create_servicechain_instance()
        self.assertRaises(webob.exc.HTTPClientError,
                          self.create_servicechain_instance)

    def test_service_profile(self):
        self.create_service_profile(service_type=constants.FIREWALL)
        self.assertRaises(webob.exc.HTTPClientError,
                          self.create_service_profile,
                          service_type=constants.FIREWALL)



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import threading

from eventlet import greenpool
from neutron_lib.db import model_base
from neutron_lib import exceptions as n_exc
from neutron_lib.plugins import constants as pconst
from oslo_config import cfg
from oslo_log import log as logging
import sqlalchemy as sa

from gbpservice._i18n import _
from gbpservice.neutron.services.servicechain.plugins.ncp import (
    exceptions as exc)
from gbpservice.nfp.common import constants as nfp_constants


NFP_NODE_DRIVER_OPTS = [
    cfg.BoolOpt('is_service_admin_owned',
                help=_("Parameter to indicate whether the Service VM has to "
                       "be owned by the Admin"),
                default=False),
    cfg.IntOpt('service_create_timeout',
               default=nfp_constants.SERVICE_CREATE_TIMEOUT,
               help=_("Seconds to wait for service creation "
                      "to complete")),
    cfg.IntOpt('service_delete_timeout',
               default=nfp_constants.SERVICE_DELETE_TIMEOUT,
               help=_("Seconds to wait for service deletion "
                      "to complete")),
]
# REVISIT(ashu): Can we use is_service_admin_owned config from RMD
cfg.CONF.register_opts(NFP_NODE_DRIVER_OPTS, "nfp_node_driver")


LOG = logging.getLogger(__name__)

# REVISIT: L2 insertion not supported
GATEWAY_PLUMBER_TYPE = [pconst.FIREWALL, pconst.VPN]
nfp_context_store = threading.local()


class InvalidServiceType(exc.NodeCompositionPluginBadRequest):
    message = _("The NFP Node driver only supports the services "
                "VPN, Firewall and LB in a Service Chain")


class ServiceProfileRequired(exc.NodeCompositionPluginBadRequest):
    message = _("A Service profile is required in Service node")


class NodeVendorMismatch(exc.NodeCompositionPluginBadRequest):
    message = _("The NFP Node driver only handles nodes which have service "
                "profile with vendor name %(vendor)s")


class DuplicateServiceTypeInChain(exc.NodeCompositionPluginBadRequest):
    message = _("The NFP Node driver does not support duplicate "
                "service types in same chain")


class RequiredProfileAttributesNotSet(exc.NodeCompositionPluginBadRequest):
    message = _("The required attributes in service profile are not present")


class InvalidNodeOrderInChain(exc.NodeCompositionPluginBadRequest):
    message = _("The NFP Node driver does not support the order "
                "of nodes defined in the current service chain spec, "
                "order should be : %(node_order)s")


class UnSupportedServiceProfile(exc.NodeCompositionPluginBadRequest):
    message = _("The NFP Node driver does not support this service "
                "profile with service type %(service_type)s and vendor "
                "%(vendor)s")


class UnSupportedInsertionMode(exc.NodeCompositionPluginBadRequest):
    message = _("The NFP Node driver supports only L3 Insertion "
                "mode")


class ServiceInfoNotAvailableOnUpdate(n_exc.NeutronException):
    message = _("Service information is not available with Service Manager "
                "on node update")


class VipNspNotSetonProvider(n_exc.NeutronException):
    message = _("Network Service policy for VIP IP address is not configured "
                "on the Providing Group")


class NodeInstanceDeleteFailed(n_exc.NeutronException):
    message = _("Node instance delete failed in NFP Node driver")


class NodeInstanceCreateFailed(n_exc.NeutronException):
    message = _("Node instance create failed in NFP Node driver")


class NodeInstanceUpdateFailed(n_exc.NeutronException):
    message = _("Node instance update failed in NFP Node driver")


class OperationNotSupported(exc.NodeCompositionPluginBadRequest):
    message = _("The NFP Node driver doesn't support operation, "
                "if instance status is in BUILD state.")


class ServiceNodeInstanceNetworkFunctionMapping(model_base.BASEV2):
    """ServiceChainInstance to NFP network function mapping."""

    __tablename__ = 'ncp_node_instance_network_function_mappings'
    sc_instance_id = sa.Column(sa.String(36),
                               nullable=False, primary_key=True)
    sc_node_id = sa.Column(sa.String(36),
                           nullable=False, primary_key=True)
    network_function_id = sa.Column(sa.String(36),
                                    nullable=True)
    status = sa.Column(sa.String(20), nullable=True)
    status_details = sa.Column(sa.String(4096), nullable=True)


class NFPContext(object):

    @staticmethod
    def store_nfp_context(sc_instance_id, **context):
        if not hasattr(nfp_context_store, 'context'):
            nfp_context_store.context = {}

        # Considering each store request comes with one entry
        if not nfp_context_store.context.get(sc_instance_id):
            NFPContext._initialise_attr(sc_instance_id)
        nfp_context_store.context[sc_instance_id].update(context)

    @staticmethod
    def clear_nfp_context(sc_instance_id):
        if not hasattr(nfp_context_store, 'context'):
            return
        if nfp_context_store.context.get(sc_instance_id):
            del nfp_context_store.context[sc_instance_id]

    @staticmethod
    def get_nfp_context(sc_instance_id):
        if not hasattr(nfp_context_store, 'context'):
            return {}
        if nfp_context_store.context.get(sc_instance_id):
            return nfp_context_store.context[sc_instance_id]
        return {}

    @staticmethod
    def _initialise_attr(sc_instance_id):
        context = {'thread_pool': greenpool.GreenPool(10),
                   'active_threads': [],
                   'sc_node_count': 0,
                   'sc_gateway_type_nodes': [],
                   'network_functions': [],
                   'update': False}
        if nfp_context_store.context:
            nfp_context_store.context.update({sc_instance_id: context})
        else:
            nfp_context_store.context = {sc_instance_id: context}



# Copyright 2014 OpenStack Foundation
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#


"""nsp_fix_constraints
"""

# revision identifiers, used by Alembic.
revision = '97ac978118c6'
down_revision = 'e8005b9b1efc'

from alembic import op
from neutron.db import migration
from sqlalchemy.engine import reflection


def upgrade(active_plugins=None, options=None):
    inspector = reflection.Inspector.from_engine(op.get_bind())
    unique_constraints = inspector.get_unique_constraints(
                                        'gp_policy_target_groups')
    op.drop_constraint('gp_policy_target_groups_ibfk_nsp',
                       'gp_policy_target_groups',
                       'foreignkey')
    for constraint in unique_constraints:
        if constraint['column_names'] == ['network_service_policy_id']:
            op.drop_constraint(constraint['name'],
                               'gp_policy_target_groups',
                               'unique')
            break
    op.create_foreign_key('gp_policy_target_groups_ibfk_nsp',
                          source_table='gp_policy_target_groups',
                          referent_table='gp_network_service_policies',
                          local_cols=['network_service_policy_id'],
                          remote_cols=['id'])
    with migration.remove_fks_from_table(
                                'gpm_service_policy_ipaddress_mappings'):
        pk_constraint = inspector.get_pk_constraint(
                'gpm_service_policy_ipaddress_mappings')
        op.drop_constraint(
                pk_constraint['name'],
                table_name='gpm_service_policy_ipaddress_mappings',
                type_='primary')
        op.create_primary_key(
                constraint_name='pk_policytargetgroup_servicepolicyid',
                table_name='gpm_service_policy_ipaddress_mappings',
                columns=['policy_target_group', 'service_policy_id'])


def downgrade(active_plugins=None, options=None):
    # Downgrade would require deleting duplicate entries resulting in data
    # loss. So skip applying back the unique constraint and change in
    # primary key
    pass



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""address_scope and subnetpool mapping for l3_policies

Revision ID: d4bb487a81b8
Revises: c1aab79622fe
Create Date: 2016-08-28 11:35:32.724952

"""

# revision identifiers, used by Alembic.
revision = 'd4bb487a81b8'
down_revision = '7afacef00d31'

from alembic import op
import sqlalchemy as sa


def upgrade():

    op.create_table(
        'gp_l3_policy_subnetpool_v4_associations',
        sa.Column('l3_policy_id', sa.String(length=36), nullable=False),
        sa.Column('subnetpool_id', sa.String(length=36), nullable=False),
        sa.ForeignKeyConstraint(['l3_policy_id'], ['gp_l3_policies.id'],
                                name='gpm_l3p_subnetpool_v4_assoc_fk_l3pid'),
        sa.ForeignKeyConstraint(['subnetpool_id'], ['subnetpools.id'],
                                name='gpm_l3p_subnetpool_v4_assoc_fk_spid'),
        sa.PrimaryKeyConstraint('l3_policy_id', 'subnetpool_id')
    )

    op.create_table(
        'gp_l3_policy_subnetpool_v6_associations',
        sa.Column('l3_policy_id', sa.String(length=36), nullable=False),
        sa.Column('subnetpool_id', sa.String(length=36), nullable=False),
        sa.ForeignKeyConstraint(['l3_policy_id'], ['gp_l3_policies.id'],
                                name='gpm_l3p_subnetpool_v6_assoc_fk_l3pid'),
        sa.ForeignKeyConstraint(['subnetpool_id'], ['subnetpools.id'],
                                name='gpm_l3p_subnetpool_v6_assoc_fk_spid'),
        sa.PrimaryKeyConstraint('l3_policy_id', 'subnetpool_id')
    )

    op.add_column(
        'gp_l3_policies',
        sa.Column('address_scope_v4_id', sa.String(length=36), nullable=True),
    )

    op.add_column(
        'gp_l3_policies',
        sa.Column('address_scope_v6_id', sa.String(length=36), nullable=True),
    )

    op.create_unique_constraint('gpm_l3p_addr_scope_v4_uq',
                                'gp_l3_policies', ['address_scope_v4_id'])
    op.create_unique_constraint('gpm_l3p_addr_scope_v6_uq',
                                'gp_l3_policies', ['address_scope_v6_id'])
    op.create_foreign_key('gpm_l3p_addr_scope_v4_fk',
                          source_table='gp_l3_policies',
                          referent_table='address_scopes',
                          local_cols=['address_scope_v4_id'],
                          remote_cols=['id'])
    op.create_foreign_key('gpm_l3p_addr_scope_v6_fk',
                          source_table='gp_l3_policies',
                          referent_table='address_scopes',
                          local_cols=['address_scope_v6_id'],
                          remote_cols=['id'])

    op.create_table(
        'gpm_owned_address_scopes',
        sa.Column('address_scope_id', sa.String(length=36), nullable=False),
        sa.ForeignKeyConstraint(
            ['address_scope_id'], ['address_scopes.id'],
            name='rmd_addr_scope_owned_fk', ondelete='CASCADE'),
        sa.PrimaryKeyConstraint('address_scope_id')
    )

    op.create_table(
        'gpm_owned_subnetpools',
        sa.Column('subnetpool_id', sa.String(length=36), nullable=False),
        sa.ForeignKeyConstraint(['subnetpool_id'], ['subnetpools.id'],
                                name='rmd_subnetpool_owned_fk',
                                ondelete='CASCADE'),
        sa.PrimaryKeyConstraint('subnetpool_id')
    )


def downgrade():
    pass



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.


import filter_base

from gbpservice.contrib.nfp.configurator.lib import data_filter


class FilterTest(filter_base.BaseTestCase):
    """Test class to test data_filter.py using unittest framework """
    def __init__(self, *args, **kwargs):
        super(FilterTest, self).__init__(*args, **kwargs)

    def setUp(self):
        """Prepare setup for every test case.
        """
        super(FilterTest, self).setUp()
        self.context = {}
        self.filter_obj = data_filter.Filter(None, None)

    def tearDown(self):
        """ Reset values after test case execution.
        """
        super(FilterTest, self).tearDown()
        self.context = {}

    def _make_test(self, context, method, **filters):
        """ To reduce the boilerplate. """
        retval = self.filter_obj.call(self.context,
                                      self.filter_obj.make_msg(method,
                                                               **filters))
        return retval

    def _make_vpn_service_context(self):
        """Make the context for the vpn service

        Returns: vpn service context

        """
        service_info = self._test_get_vpn_info()
        self.context['service_info'] = service_info
        return self.context

    def _make_fw_service_context(self):
        """Make the context for the fw service

        Returns: fw service context

        """
        service_info = self._test_get_fw_info()
        self.context['service_info'] = service_info
        return self.context

    def test_get_vpn_service_with_tenantid(self):
        """Test get_vpn_services() of data_filter.py by passing
           only tenant_id in filters
        """
        retval = self._make_test(self._make_vpn_service_context(),
                                 'get_vpn_services',
                                 filters=(
                            {'tenant_id': [self.vpnservices[0]['tenant_id']]}))

        self.assertEqual(retval, [self.vpnservices[0], self.vpnservices[1]])

    def test_get_vpn_service_with_ids(self):
        """Test get_vpn_services() of data_filter.py by passing
           vpn service ids in filters
        """
        retval = self._make_test(self._make_vpn_service_context(),
                                 'get_vpn_services',
                                 ids=[self.vpnservices[0]['id'],
                                      self.vpnservices[1]['id']])
        self.assertEqual(retval, [self.vpnservices[0], self.vpnservices[1]])

    def test_get_ipsec_conns(self):
        """Test get_ipsec_conns() of data_filter.py
        """
        retval = self._make_test(
                self._make_vpn_service_context(),
                'get_ipsec_conns',
                tenant_id=[self.ipsec_site_connections[0]['tenant_id']],
                peer_address=[self.ipsec_site_connections[0]['peer_address']])
        self.assertEqual(retval, self.ipsec_site_connections)

    def test_get_vpn_servicecontext_ipsec_service_type(self):
        """Test get_vpn_servicecontext() of data_filter.py
           based on ipsec service type
        """
        service_info = self._test_get_vpn_info()
        self.context['service_info'] = service_info
        retval = self.filter_obj._get_vpn_servicecontext(
                    self.context,
                    {'tenant_id': self.vpnservices[0]['tenant_id'],
                     'vpnservice_id': self.vpnservices[0]['id'],
                     'ipsec_site_connections':
                     self.ipsec_site_connections[0]['id']})

        expected = {'service': self.vpnservices[0],
                    'siteconns': [{'connection':
                                   self.ipsec_site_connections[0],
                                   'ikepolicy': self.ikepolicies[0],
                                   'ipsecpolicy': self.ipsecpolicies[0]
                                   }]}

        self.assertEqual(retval, [expected])

    def test_get_vpn_servicecontext_ipsec_service_type_with_tenantid(self):
        """Test get_vpn_servicecontext() of data_filter.py
           based on ipsec service type and tenant_id
        """
        service_info = self._test_get_vpn_info()
        self.context['service_info'] = service_info
        retval = self.filter_obj._get_vpn_servicecontext(
                    self.context,
                    {'tenant_id': self.vpnservices[0]['tenant_id'],
                     })

        expected = {'service': self.vpnservices[0],
                    'siteconns': [{'connection':
                                   self.ipsec_site_connections[0],
                                   'ikepolicy': self.ikepolicies[0],
                                   'ipsecpolicy': self.ipsecpolicies[0]
                                   }]}

        self.assertEqual(retval, [expected])



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from unittest import mock

from mock import call
import webob.exc

from oslo_config import cfg
from vmware_nsx.common import config
from vmware_nsxlib.v3 import exceptions as nsxlib_exc

from gbpservice.neutron.services.grouppolicy.drivers.vmware.nsx_policy import (
    nsx_policy_mapping as driver)
from gbpservice.neutron.tests.unit.services.grouppolicy import (
    test_resource_mapping as test_rmd)


TEST_PROJECT = 'test-project'


class NsxPolicyMappingTestCase(test_rmd.ResourceMappingTestCase):

    def setUp(self):
        self.set_up_mocks()
        self.set_up_config()

        super(NsxPolicyMappingTestCase, self).setUp(
            policy_drivers=['implicit_policy', 'nsx_policy'])
        # REVISIT (annak): currently run with ML2 plugin
        # core_plugin='vmware_nsx.plugin.NsxV3Plugin'

        self.driver = self._gbp_plugin.policy_driver_manager.policy_drivers[
               'nsx_policy'].obj
        self.nsx_policy = self.driver.nsx_policy
        self.nsx_port = self.driver.nsx_port
        self._tenant_id = TEST_PROJECT

    def tearDown(self):
        super(NsxPolicyMappingTestCase, self).tearDown()
        self.patcher.stop()

    def set_up_config(self):
        cfg.CONF.register_opts(driver.policy_opts, driver.DRIVER_OPT_GROUP)
        cfg.CONF.register_opts(config.nsx_v3_opts, group="nsx_v3")
        cfg.CONF.set_override('nsx_policy_manager', '1.1.1.1',
                              driver.DRIVER_OPT_GROUP)
        cfg.CONF.set_override('nsx_api_managers', '1.1.1.1',
                              driver.NSX_V3_GROUP)

    def set_up_mocks(self):
        mock.patch("vmware_nsxlib.v3.client.NSX3Client").start()
        mock.patch("vmware_nsxlib.v3.policy_resources"
                   ".NsxPolicyEnforcementPointApi").start()
        mock.patch("vmware_nsxlib.v3.cluster.ClusteredAPI"
                   "._init_endpoints").start()
        # To avoid:
        # '>' not supported between instances of 'MagicMock' and 'int'
        # due to
        # `MAX_NSGROUPS_CRITERIA_TAGS = max(10, max_tags - 5)`
        # in vmware_nsxlib.v3.utils._update_max_nsgroups_criteria_tags()
        self.patcher = mock.patch("vmware_nsxlib.v3.utils."
            "_update_max_nsgroups_criteria_tags")
        self.patcher.return_value = 10
        self.patcher.start()

    def _mock_domain_create(self):
        return mock.patch.object(self.nsx_policy.domain, 'create_or_overwrite')

    def _mock_domain_delete(self):
        return mock.patch.object(self.nsx_policy.domain, 'delete')

    def _mock_service_create(self):
        return mock.patch.object(self.nsx_policy.service,
                                 'create_or_overwrite')

    def _mock_icmp_service_create(self):
        return mock.patch.object(self.nsx_policy.icmp_service,
                                 'create_or_overwrite')

    def _mock_service_delete(self):
        return mock.patch.object(self.nsx_policy.service, 'delete')

    def _mock_icmp_service_delete(self):
        return mock.patch.object(self.nsx_policy.icmp_service, 'delete')

    def _mock_group_create(self):
        return mock.patch.object(self.nsx_policy.group, 'create_or_overwrite')

    def _mock_group_create_fails(self):
        return mock.patch.object(self.nsx_policy.group, 'create_or_overwrite',
                                 side_effect=nsxlib_exc.ManagerError)

    def _mock_group_delete(self):
        return mock.patch.object(self.nsx_policy.group, 'delete')

    def _mock_map_create(self):
        return mock.patch.object(self.nsx_policy.comm_map,
                                 'create_or_overwrite')

    def _mock_map_delete(self):
        return mock.patch.object(self.nsx_policy.comm_map, 'delete')

    def _mock_map_create_fails(self):
        return mock.patch.object(self.nsx_policy.comm_map,
                                 'create_or_overwrite',
                                 side_effect=nsxlib_exc.ManagerError)

    def _mock_nth_map_create_fails(self, n=2):
        self.call_count = 1

        def raise_on_nth_call(**kwargs):
            if self.call_count == n:
                raise nsxlib_exc.ManagerError
            else:
                self.call_count += 1
        return mock.patch.object(self.nsx_policy.comm_map,
                                 'create_or_overwrite',
                                 side_effect=raise_on_nth_call)

    def _mock_policy_create_fails(self):
        return mock.patch.object(self.policy_api, 'create_or_overwrite',
                                 side_effect=nsxlib_exc.ManagerError)

    def _mock_policy_delete(self):
        return mock.patch.object(self.policy_api, 'delete')

    def _mock_nsx_db(self):
        def mirror_port_id(session, port_id):
            return None, port_id
        mock.patch('vmware_nsx.db.db.get_nsx_switch_and_port_id',
                   side_effect=mirror_port_id).start()

    def _mock_nsx_port_update(self):
        return mock.patch.object(self.nsx_port, 'update')


class TestPolicyClassifier(NsxPolicyMappingTestCase):

    def test_l4_lifecycle(self):
        with self._mock_service_create() as service_create_call, \
            self._mock_service_delete() as service_delete_call:

            # classifier create
            cl = self.create_policy_classifier(
                name='test',
                protocol='TCP',
                port_range='80',
                direction='bi')['policy_classifier']

            # verify API call to create the service
            service_create_call.assert_called_with(
                name=mock.ANY,
                description=mock.ANY,
                protocol='tcp',
                dest_ports=['80'],
                service_id=mock.ANY)

            service_create_call.reset_mock()

            # classifier update
            cl = self.update_policy_classifier(
                    cl['id'],
                    port_range='443')['policy_classifier']

            service_create_call.assert_called_with(
                name=mock.ANY,
                description=mock.ANY,
                protocol='tcp',
                dest_ports=['443'],
                service_id=mock.ANY)

            # classifier delete
            self.delete_policy_classifier(cl['id'])

            service_delete_call.assert_called_with(cl['id'])

    def test_create_port_range(self):
        with self._mock_service_create() as service_create_call:

            self.create_policy_classifier(name='test',
                                          protocol='UDP',
                                          port_range='777:888',
                                          direction='in')

            port_list = [str(p) for p in range(777, 889)]
            service_create_call.assert_called_with(
                name=mock.ANY,
                description=mock.ANY,
                protocol='udp',
                dest_ports=port_list,
                service_id=mock.ANY)

    def test_create_without_ports(self):
        with self._mock_service_create() as service_create_call:

            self.create_policy_classifier(name='test',
                                          protocol='TCP',
                                          direction='in')

            service_create_call.assert_called_with(
                name=mock.ANY,
                description=mock.ANY,
                protocol='tcp',
                dest_ports=[],
                service_id=mock.ANY)

    def test_icmp_lifecycle(self):
        with self._mock_icmp_service_create() as service_create_call, \
            self._mock_icmp_service_delete() as service_delete_call:

            cl = self.create_policy_classifier(
                name='test',
                protocol='icmp',
                direction='bi')['policy_classifier']

            # verify API call to create the service
            service_create_call.assert_called()

            self.delete_policy_classifier(cl['id'])

            service_delete_call.assert_called_with(cl['id'])

    def test_update_protocol_fails(self):
        with self._mock_icmp_service_create():

            cl = self.create_policy_classifier(
                name='test',
                protocol='icmp',
                direction='bi')['policy_classifier']

            self.assertRaises(webob.exc.HTTPClientError,
                              self.update_policy_classifier,
                              cl['id'],
                              protocol='tcp',
                              dest_ports=['80'])

    def test_icmpv6_protocol_fails(self):
        self.assertRaises(webob.exc.HTTPClientError,
                          self.create_policy_classifier,
                          name='test',
                          protocol='58',
                          direction='bi')


class TestPolicyTargetGroup(NsxPolicyMappingTestCase):

    def _prepare_rule_set(self, name='test_rule',
                          direction='bi', rule_count=1):
        self.rules = []
        with self._mock_service_create():

            for i in range(rule_count):
                self.rules.append(
                    self._create_simple_policy_rule(direction=direction))

            rule_ids = [rule['id'] for rule in self.rules]
            return self.create_policy_rule_set(name=name,
                policy_rules=rule_ids)['policy_rule_set']

    def assert_neutron_resources(self, net_count, subnet_count, port_count):
        networks = self._plugin.get_networks(self._context)
        self.assertEqual(net_count, len(networks))

        subnets = self._plugin.get_subnets(self._context)
        self.assertEqual(subnet_count, len(subnets))

        ports = self._plugin.get_ports(self._context)
        self.assertEqual(port_count, len(ports))

    def assert_neutron_rollback(self):
        self.assert_neutron_resources(0, 0, 0)

    def group_call(self, name, group_id):
        return call(domain_id=TEST_PROJECT,
                    name=driver.generate_nsx_name(group_id, name),
                    description=mock.ANY,
                    cond_val=group_id,
                    group_id=group_id)

    def ingress_map_call(self, prs_id, provider_ids,
                         consumer_ids, service_ids):
        return call(domain_id=TEST_PROJECT,
                    map_id=mock.ANY,
                    name=driver.append_in_dir(prs_id),
                    service_ids=service_ids,
                    description=mock.ANY,
                    source_groups=consumer_ids,
                    dest_groups=provider_ids,
                    )

    def egress_map_call(self, prs_id, provider_ids,
                        consumer_ids, service_ids):
        return call(domain_id=TEST_PROJECT,
                    map_id=mock.ANY,
                    service_ids=service_ids,
                    name=driver.append_out_dir(prs_id),
                    description=mock.ANY,
                    source_groups=provider_ids,
                    dest_groups=consumer_ids)

    def test_create_first_ptg_for_project(self):
        '''Create first ptg for tenant and verify domain creation'''

        with self._mock_domain_create() as domain_create,\
            self._mock_group_create() as group_create,\
            self._mock_map_create() as map_create:

            ptg = self.create_policy_target_group(
                name='test')['policy_target_group']

            domain_create.assert_called_with(domain_id=TEST_PROJECT,
                                             name=mock.ANY,
                                             description=mock.ANY)
            group_create.assert_has_calls([self.group_call('test', ptg['id'])])
            map_create.assert_not_called()

    def _test_ptg_pair_with_rule_set(self,
                                     direction_in=True,
                                     direction_out=True,
                                     rule_count=1):
        '''Test consumer and producer group pair with single rule lifecycle.

        Verify backend group and rule creation calls.
        Verify spawned neutron resources.
        '''

        direction = 'bi'
        if not direction_in:
            direction = 'out'
        if not direction_out:
            direction = 'in'

        policy_rule_set = self._prepare_rule_set(direction=direction,
                                                 rule_count=rule_count)

        # Create group pair
        with self._mock_group_create() as group_create,\
            self._mock_map_create() as map_create,\
            self._mock_domain_create():

            provider_ptg, consumer_ptg = self._create_provider_consumer_ptgs(
                    policy_rule_set['id'])

            # validate group creation on backend
            calls = [self.group_call('ptg1', provider_ptg),
                     self.group_call('ptg2', consumer_ptg)]
            group_create.assert_has_calls(calls)

            # validate communication map creation on backend
            calls = []
            classifier_ids = sorted([rule['policy_classifier_id']
                                     for rule in self.rules])
            if direction_in:
                calls.append(self.ingress_map_call(
                    policy_rule_set['id'],
                    [provider_ptg],
                    [consumer_ptg],
                    classifier_ids))
            if direction_out:
                calls.append(self.egress_map_call(
                    policy_rule_set['id'],
                    [provider_ptg],
                    [consumer_ptg],
                    classifier_ids))
            map_create.assert_has_calls(calls)

            # validate neutron resources
            self.assert_neutron_resources(2, 2, 2)

        # Delete producer
        with self._mock_map_delete() as map_delete,\
            self._mock_group_delete() as group_delete,\
            self._mock_domain_delete() as domain_delete:

            self.delete_policy_target_group(provider_ptg)

            # verify communication map delete on backend
            calls = []
            if direction_in:
                calls.append(call(
                    TEST_PROJECT,
                    driver.append_in_dir(policy_rule_set['id'])))
            if direction_out:
                calls.append(call(
                    TEST_PROJECT,
                    driver.append_out_dir(policy_rule_set['id'])))

            map_delete.assert_has_calls(calls)

            # verify group delete call
            group_delete.assert_called_with(TEST_PROJECT, provider_ptg)

            # verify domain not deleted yet
            domain_delete.assert_not_called()

        # Delete consumer
        with self._mock_map_delete() as map_delete,\
            self._mock_group_delete() as group_delete,\
            self._mock_domain_delete() as domain_delete:

            self.delete_policy_target_group(consumer_ptg)

            # no deletions on communication map are expected
            map_delete.assert_not_called()

            # verify group delete call
            group_delete.assert_called_with(TEST_PROJECT, consumer_ptg)

            # last group is deleted, domain should go as well
            domain_delete.assert_called_with(TEST_PROJECT)

    def test_ptg_pair_with_single_rule_in(self):
        self._test_ptg_pair_with_rule_set(True, False)

    def test_ptg_pair_with_single_rule_out(self):
        self._test_ptg_pair_with_rule_set(False, True)

    def test_ptg_pair_with_single_rule_bi(self):
        self._test_ptg_pair_with_rule_set(True, True)

    def test_ptg_pair_with_multi_rule_in(self):
        self._test_ptg_pair_with_rule_set(True, False, 2)

    def test_ptg_pair_with_multi_rule_out(self):
        self._test_ptg_pair_with_rule_set(False, True, 3)

    def test_ptg_pair_with_multi_rule_bi(self):
        self._test_ptg_pair_with_rule_set(True, True, 5)

    def test_create_fail_isolated(self):
        '''Verify integrity when backend fails on isolated group creation.

        Verify backend receives a group delete call.
        Verify spawned neutron resources are cleaned up.
        '''

        policy_rule_set = self._prepare_rule_set()

        with self._mock_domain_create(),\
            self._mock_group_create_fails(),\
            self._mock_group_delete() as group_delete,\
            self._mock_domain_delete() as domain_delete:

            self.assertRaises(webob.exc.HTTPClientError,
                              self._create_provider_consumer_ptgs,
                              policy_rule_set['id'])

            group_delete.assert_called_with(self._tenant_id,
                                            mock.ANY)

            # verify domain deletion since group failed to create
            domain_delete.assert_called_with(TEST_PROJECT)

            self.assert_neutron_rollback()

    def test_create_fail_connected(self):
        '''Verify integrity when backend fails on connectivity map creation

        This test creates a pair of groups. First group creation succeeds,
        while second fails on connectivity enforcement.
        Verify backend receives a group delete call for second group.
        Verify spawned neutron resources are cleaned up for second group.
        '''

        policy_rule_set = self._prepare_rule_set()

        with self._mock_group_create(),\
            self._mock_map_create_fails(),\
            self._mock_group_delete() as group_delete:

            self.assertRaises(webob.exc.HTTPClientError,
                              self._create_provider_consumer_ptgs,
                              policy_rule_set['id'])

            group_delete.assert_called_with(self._tenant_id, mock.ANY)

            self.assert_neutron_resources(1, 1, 1)

    def test_create_fail_multi_connected(self):
        '''Verify integrity when backend fails on connectivity map creation

        This test creates three groups a<-->b<==>c
        B is created last, and creation fails on its last connectivity
        enforcement.
        Verify all maps are deleted in cleanup.
        Verify spawned neutron resources are cleaned up for third group.
        '''

        prs1 = self._prepare_rule_set()['id']
        service1 = self.rules[0]['policy_classifier_id']
        prs2 = self._prepare_rule_set()['id']
        service2 = self.rules[0]['policy_classifier_id']
        prs3 = self._prepare_rule_set()['id']
        service3 = self.rules[0]['policy_classifier_id']

        # Create a and c
        with self._mock_group_create(),\
            self._mock_map_create():

            ab_dict = {prs1: None}
            bc_dict = {prs2: None, prs3: None}
            a = self.create_policy_target_group(
                name='a',
                provided_policy_rule_sets=ab_dict)['policy_target_group']['id']
            c = self.create_policy_target_group(
                name='c',
                consumed_policy_rule_sets=bc_dict)['policy_target_group']['id']

        with self._mock_group_create(),\
            self._mock_nth_map_create_fails(n=6) as map_create,\
            self._mock_map_delete() as map_delete,\
            self._mock_group_delete() as group_delete:

            self.assertRaises(webob.exc.HTTPClientError,
                              self.create_policy_target_group,
                              name='c',
                              consumed_policy_rule_sets=ab_dict,
                              provided_policy_rule_sets=bc_dict)

            b = mock.ANY
            map_create_calls = [
                self.ingress_map_call(prs1, [a], [b], [service1]),
                self.egress_map_call(prs1, [a], [b], [service1]),
                self.ingress_map_call(prs2, [b], [c], [service2]),
                self.egress_map_call(prs2, [b], [c], [service2]),
                self.ingress_map_call(prs3, [b], [c], [service3]),
                self.egress_map_call(prs3, [b], [c], [service3])]

            map_create.assert_has_calls(map_create_calls, any_order=True)

            map_delete_calls = [
                call(TEST_PROJECT, driver.append_in_dir(prs1)),
                call(TEST_PROJECT, driver.append_out_dir(prs1)),
                call(TEST_PROJECT, driver.append_in_dir(prs2)),
                call(TEST_PROJECT, driver.append_out_dir(prs2)),
                call(TEST_PROJECT, driver.append_in_dir(prs3))]

            map_delete.assert_has_calls(map_delete_calls, any_order=True)

            group_delete.assert_called_with(TEST_PROJECT, mock.ANY)

            self.assert_neutron_resources(2, 2, 2)

    def test_create_ptg_pair_multi_rule_set(self):
        '''Create ptg pair based on 3 rule sets

        First rule set is simulated to have only ingress connectivity,
        second - only egress connectivity, and third - both
        '''
        prs1 = self._prepare_rule_set()['id']
        service1 = self.rules[0]['policy_classifier_id']
        prs2 = self._prepare_rule_set()['id']
        service2 = self.rules[0]['policy_classifier_id']
        prs3 = self._prepare_rule_set()['id']
        service3 = self.rules[0]['policy_classifier_id']

        with self._mock_domain_create(),\
            self._mock_group_create() as group_create,\
            self._mock_map_create() as map_create:

            rule_set_dict = {prs1: None, prs2: None, prs3: None}
            provider_ptg = self.create_policy_target_group(
                name='ptg1', provided_policy_rule_sets=rule_set_dict)
            provider_id = provider_ptg['policy_target_group']['id']
            consumer_ptg = self.create_policy_target_group(
                name='ptg2', consumed_policy_rule_sets=rule_set_dict)
            consumer_id = consumer_ptg['policy_target_group']['id']

            group_create.assert_has_calls(
                [self.group_call('ptg1', provider_id),
                 self.group_call('ptg2', consumer_id)])

            map_calls = [
                self.ingress_map_call(prs1, [provider_id],
                                      [consumer_id], [service1]),
                self.egress_map_call(prs2, [provider_id],
                                     [consumer_id], [service2]),
                self.ingress_map_call(prs3, [provider_id],
                                      [consumer_id], [service3]),
                self.egress_map_call(prs3, [provider_id],
                                     [consumer_id], [service3])]
            map_create.assert_has_calls(map_calls, any_order=True)

    def test_create_ptg_ring(self):
        ring_size = 10

        prs_ids = []
        services = []
        for i in range(0, ring_size):
            prs_ids.append(self._prepare_rule_set()['id'])
            services.append(self.rules[0]['policy_classifier_id'])

        # Create ring topology
        with self._mock_domain_create(),\
            self._mock_group_create() as group_create,\
            self._mock_map_create() as map_create:

            group_calls = []
            map_calls = []
            ptg_ids = []
            for i in range(0, ring_size):
                provided_rule_set_dict = {prs_ids[i]: None}
                next_i = (i + 1) % ring_size
                consumed_rule_set_dict = {prs_ids[next_i]: None}
                name = 'ptg_%d' % i
                ptg = self.create_policy_target_group(
                        name=name,
                        provided_policy_rule_sets=provided_rule_set_dict,
                        consumed_policy_rule_sets=consumed_rule_set_dict)

                ptg_id = ptg['policy_target_group']['id']
                ptg_ids.append(ptg_id)
                group_calls.append(self.group_call(name, ptg_id))

                if i > 0:
                    map_calls.append(self.ingress_map_call(
                        prs_ids[i],
                        [ptg_id],
                        [ptg_ids[i - 1]],
                        [services[i]]))

            map_calls.append(self.ingress_map_call(prs_ids[0],
                                                   [ptg_ids[0]],
                                                   [ptg_id],
                                                   [services[0]]))

            group_create.assert_has_calls(group_calls)
            map_create.assert_has_calls(map_calls, any_order=True)

            self.assert_neutron_resources(ring_size, ring_size, ring_size)

        # Delete single group and verify connectors are deleted
        with self._mock_map_delete() as map_delete,\
            self._mock_map_create() as map_create,\
            self._mock_group_delete() as group_delete:

            ptg_id = ptg_ids[2]
            self.delete_policy_target_group(ptg_id)

            map_calls = [call(TEST_PROJECT, driver.append_in_dir(prs_ids[2])),
                         call(TEST_PROJECT, driver.append_out_dir(prs_ids[2])),
                         call(TEST_PROJECT, driver.append_in_dir(prs_ids[3])),
                         call(TEST_PROJECT, driver.append_out_dir(prs_ids[3]))]

            map_delete.assert_has_calls(map_calls)
            map_create.assert_not_called()
            group_delete.assert_called_with(TEST_PROJECT, ptg_id)

        # Remove connectors from single group
        with self._mock_map_delete() as map_delete,\
            self._mock_map_create() as map_create,\
            self._mock_group_delete() as group_delete:

            ptg_id = ptg_ids[5]
            self.update_policy_target_group(
                ptg_id, provided_policy_rule_sets={})

            map_calls = [call(TEST_PROJECT, driver.append_in_dir(prs_ids[5]))]
            map_delete.assert_has_calls(map_calls)
            map_create.assert_not_called()
            group_delete.assert_not_called()

    def test_create_ptg_star(self):
        '''Star-like topology (single producer and N consumers) lifecycle'''

        star_size = 10
        policy_rule_set = self._prepare_rule_set()
        service = self.rules[0]['policy_classifier_id']
        prs_id = policy_rule_set['id']

        # Create topology
        with self._mock_domain_create(),\
            self._mock_group_create() as group_create,\
            self._mock_map_create() as map_create:

            policy_rule_set_dict = {prs_id: None}
            provider_ptg = self.create_policy_target_group(
                    name='producer',
                    provided_policy_rule_sets=policy_rule_set_dict)
            provider_id = provider_ptg['policy_target_group']['id']

            group_calls = [self.group_call('producer', provider_id)]
            map_calls = []

            consumer_ids = []
            for i in range(0, star_size):
                name = 'consumer_%d' % i
                consumer_ptg = self.create_policy_target_group(
                        name=name,
                        consumed_policy_rule_sets=policy_rule_set_dict)

                consumer_id = consumer_ptg['policy_target_group']['id']
                consumer_ids.append(consumer_id)

                group_calls.append(self.group_call(name, consumer_id))

                map_calls.append(self.ingress_map_call(
                    prs_id,
                    [provider_id],
                    consumer_ids[:],
                    [service]))

                map_calls.append(self.egress_map_call(
                    prs_id,
                    [provider_id],
                    consumer_ids[:],
                    [service]))

            group_create.assert_has_calls(group_calls)
            map_create.assert_has_calls(map_calls)

            star_size += 1
            self.assert_neutron_resources(star_size, star_size, star_size)

        # Delete one consumer group
        with self._mock_map_delete() as map_delete,\
            self._mock_map_create() as map_create,\
            self._mock_group_delete() as group_delete:

            consumer_id = consumer_ids.pop(0)
            self.delete_policy_target_group(consumer_id)

            map_create.assert_has_calls(
                [self.ingress_map_call(prs_id,
                                      [provider_id],
                                      consumer_ids,
                                      [service])])

            map_delete.assert_not_called()

            group_delete.assert_called_with(TEST_PROJECT, consumer_id)

            star_size -= 1
            self.assert_neutron_resources(star_size, star_size, star_size)

        # Delete provider group
        with self._mock_map_delete() as map_delete,\
            self._mock_map_create() as map_create,\
            self._mock_group_delete() as group_delete:

            self.delete_policy_target_group(provider_id)

            map_create.assert_not_called()
            map_delete.assert_has_calls(
                [call(TEST_PROJECT, driver.append_in_dir(prs_id)),
                 call(TEST_PROJECT, driver.append_out_dir(prs_id))])

            star_size -= 1
            group_delete.assert_called_with(TEST_PROJECT, provider_id)

    def test_update_rule_set(self):
        '''Under consumer and producer group pair with single rule lifecycle,
        Update rule set and verify backend gets updated
        '''
        rule_set = self._prepare_rule_set(direction='in')

        # Create group pair
        with self._mock_group_create(), \
            self._mock_map_create(), \
            self._mock_domain_create():

            provider_ptg, consumer_ptg = self._create_provider_consumer_ptgs(
                    rule_set['id'])

        # Add another rule to the rule set
        with self._mock_service_create(),\
            self._mock_map_create() as map_update:

            rule = self._create_simple_policy_rule(direction='in')
            self.rules.append(rule)

            rules = [rule['id'] for rule in self.rules]
            self.update_policy_rule_set(rule_set['id'], policy_rules=rules)
            classifier_ids = sorted([rule['policy_classifier_id']
                                     for rule in self.rules])

            map_update.assert_has_calls([self.ingress_map_call(
                rule_set['id'],
                [provider_ptg],
                [consumer_ptg],
                classifier_ids)])

        # Delete all rules from rule set, leaving the rule set
        with self._mock_map_delete() as map_delete:
            self.update_policy_rule_set(rule_set['id'], policy_rules=[])

            # verify communication map delete on backend
            map_delete.assert_any_call(
                TEST_PROJECT, driver.append_in_dir(rule_set['id']))

    def test_update_rule_classifier(self):
        '''Under consumer and producer group pair with single rule lifecycle,
        Update rule set and verify backend gets updated
        '''
        rule_set = self._prepare_rule_set(direction='in')

        # Create group pair
        with self._mock_group_create(), \
            self._mock_map_create(), \
            self._mock_domain_create():

            provider_ptg, consumer_ptg = self._create_provider_consumer_ptgs(
                    rule_set['id'])

        # Update classifier in rule
        with self._mock_service_create(),\
            self._mock_map_create() as map_update:

            cl = self.create_policy_classifier(
                name='test',
                protocol='UDP',
                port_range='512',
                direction='bi')['policy_classifier']

            self.update_policy_rule(self.rules[0]['id'],
                                    policy_classifier_id=cl['id'])

            map_update.assert_has_calls([self.ingress_map_call(
                rule_set['id'],
                [provider_ptg],
                [consumer_ptg],
                [cl['id']])])


class TestPolicyRuleSet(NsxPolicyMappingTestCase):

    def test_basic(self):
        ''' Create and delete unused rule set'''

        rule = self._create_simple_policy_rule()
        rule_set = self.create_policy_rule_set(name='test',
            policy_rules=[rule['id']])['policy_rule_set']

        self.update_policy_rule_set(rule_set['id'], name='test1')
        self.delete_policy_rule_set(rule_set['id'])


class TestPolicyTargetTag(NsxPolicyMappingTestCase):

    def _prepare_group(self, name='test'):
        with self._mock_group_create():
            return self.create_policy_target_group(
                name='test')['policy_target_group']

    def test_target_lifecycle(self):
        self._mock_nsx_db()

        ptg = self._prepare_group()

        # create policy target and verify port tag update
        with self._mock_nsx_port_update() as port_update:

            target = self.create_policy_target(
                policy_target_group_id=ptg['id'])['policy_target']

            # nsx mock function will map neutron port id to same value
            # for nsx port id
            port_update.assert_called_once_with(
                target['port_id'],
                None,
                tags_update=[{'scope': 'gbp',
                              'tag': ptg['id']}])

        # verify group membership change is not supported
        ptg1 = self._prepare_group()
        self.assertRaises(webob.exc.HTTPClientError,
                          self.update_policy_target,
                          target['id'],
                          policy_target_group_id=ptg1['id'])

        # policy target deletion should not affect backend policy-wise
        self.delete_policy_target(target['id'])


class TestL3Policy(NsxPolicyMappingTestCase):

    def test_ipv6_supported(self):
        self.assertRaises(webob.exc.HTTPClientError,
                          self.create_l3_policy,
                          ip_version=6,
                          ip_pool='1001::0/64')
